# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.
#
# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
from multiprocessing.managers import ListProxy
import time

# Third Party Imports
import numpy as np
import nidaqmx

# Local Imports
from navigate.model.devices.stages.stage_base import StageBase
from navigate.model.waveforms import dc_value, remote_focus_ramp

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class GalvoNIStage(StageBase):
    """Galvo Stage Class (only supports one axis)

    Generic analog controlled stage. Could be used to control piezos, galvos,
    etc. Currently set up to handle National Instruments data acquisition cards.

    Retrieves the volts per micron from the configuration file and uses that to
    determine the correct voltage to send to the stage.

    """

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the Galvo Stage.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocessing.managers.DictProxy
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration, device_id)

        # 1 V/ 100 um
        device_config = configuration["configuration"]["microscopes"][microscope_name][
            "stage"
        ]["hardware"]

        if type(device_config) == ListProxy:
            #: float: volts per micron scaling factor.
            self.volts_per_micron = device_config[device_id]["volts_per_micron"]

            #: int: channel number for each axis
            self.axes_channels = device_config[device_id]["axes_mapping"]

            #: float: maximum voltage for each axis
            self.galvo_max_voltage = device_config[device_id]["max"]

            #: float: minimum voltage for each axis
            self.galvo_min_voltage = device_config[device_id]["min"]

            #: float: Distance threshold for wait until delay
            self.distance_threshold = device_config[device_id].get(
                "distance_threshold", None
            )  # microns

            #: float: Stage settle duration in milliseconds.
            self.stage_settle_duration = (
                device_config[device_id].get("settle_duration_ms", 20) / 1000
            )  # convert to seconds
        else:
            self.volts_per_micron = device_config["volts_per_micron"]
            self.axes_channels = device_config["axes_mapping"]
            self.galvo_max_voltage = device_config["max"]
            self.galvo_min_voltage = device_config["min"]
            self.distance_threshold = device_config.get("distance_threshold", None)
            self.stage_settle_duration = (
                device_config.get("settle_duration_ms", 20) / 1000
            )

        #: dict: Mapping of software axes to hardware axes.
        self.axes_mapping = {self.axes[0]: self.axes_channels[0]}

        #: object: Hardware device to connect to.
        self.daq = device_connection

        #: str: Name of the microscope.
        self.microscope_name = microscope_name

        #: multiprocessing.managers.DictProxy: Global configuration of the microscope.
        self.configuration = configuration

        #: str: Trigger source for the DAQ.
        self.trigger_source = configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["trigger_source"]

        #: float: Percent of the camera delay.
        self.camera_delay_percent = configuration["configuration"]["microscopes"][
            microscope_name
        ]["camera"]["delay_percent"]

        #: float: Percent of the remote focus delay.
        self.remote_focus_ramp_falling = configuration["configuration"]["microscopes"][
            microscope_name
        ]["remote_focus_device"]["ramp_falling_percent"]

        #: float: Percent of the remote focus delay.
        self.remote_focus_delay = configuration["configuration"]["microscopes"][
            microscope_name
        ]["remote_focus_device"]["delay_percent"]

        #: float: Sample rate of the DAQ.
        self.sample_rate = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["daq"]["sample_rate"]

        #: dict: Dictionary of exposure times for each channel.
        self.exposure_times = None

        #: dict: Dictionary of sweep times for each channel.
        self.sweep_times = None

        #: dict: Dictionary of waveforms for each channel.
        self.waveform_dict = {}

        #: object: DAQ ao task
        self.ao_task = None

        self.switch_mode("normal")

    # for stacking, we could have 2 axis here or not, y is for tiling, not necessary
    def report_position(self):
        """Reports the position for all axes, and create position dictionary.

        Returns
        -------
        dict
            Dictionary of positions for each axis.
        """
        return self.get_position_dict()
    
    def update_waveform(self, waveform_dict):
        """Update the waveform for the stage.

        Parameters
        ----------
        waveform_dict : dict
            Dictionary of waveforms for each channel

        Returns
        -------
        result : bool
            success or failed
        """
        self.switch_mode("waveform")
        microscope_state = self.configuration["experiment"]["MicroscopeState"]
        
        for channel_key in microscope_state["channels"].keys():
            # channel includes 'is_selected', 'laser', 'filter', 'camera_exposure'...
            channel = microscope_state["channels"][channel_key]

            # Only proceed if it is enabled in the GUI
            if channel["is_selected"] is True:

                if channel_key not in waveform_dict:
                    logger.debug(f"Update waveform in StageGalvo failed! {channel_key}")
                    print("*** updating waveform in StageGalvo failed!", channel_key)
                    return False

                waveform_dict[channel_key][
                    waveform_dict[channel_key] > self.galvo_max_voltage
                ] = self.galvo_max_voltage
                waveform_dict[channel_key][
                    waveform_dict[channel_key] < self.galvo_min_voltage
                ] = self.galvo_min_voltage

        self.waveform_dict = waveform_dict
        self.daq.analog_outputs[self.axes_channels[0]] = {
            "trigger_source": self.trigger_source,
            "waveform": self.waveform_dict,
        }
        return True

    def move_axis_absolute(self, axis, abs_pos, wait_until_done=False):
        """Implement movement logic along a single axis.

        Parameters
        ----------
        axis : str
            An axis prefix in move_dictionary.
            For example, axis='x' corresponds to 'x_abs', 'x_min', etc.
        abs_pos : float
            Absolute position value
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """
        self.waveform_dict = dict.fromkeys(self.waveform_dict, None)
        axis_abs = self.get_abs_position(axis, abs_pos)
        if axis_abs == -1e50:
            return False

        # Keep track of step size.
        current_position = getattr(self, f"{axis}_pos", axis_abs)
        delta_position = np.abs(axis_abs - current_position)

        volts = eval(self.volts_per_micron, {"x": axis_abs})
        if volts > self.galvo_max_voltage:
            volts = self.galvo_max_voltage
        if volts < self.galvo_min_voltage:
            volts = self.galvo_min_voltage

        self.ao_task.write(volts, auto_start=True)
        # Stage Settle Duration in Milliseconds
        if (
            wait_until_done
            and (self.distance_threshold is not None)
            and (delta_position >= self.distance_threshold)
        ):
            time.sleep(self.stage_settle_duration)

        setattr(self, f"{axis}_pos", axis_abs)

        return True

    def move_absolute(self, move_dictionary, wait_until_done=True):
        """Move Absolute Method.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement.
            Includes 'x_abs', etc. for one or more axes.
            Expects values in micrometers, except for theta, which is in degrees.
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        success : bool
            Was the move successful?
        """
        abs_pos_dict = self.verify_abs_position(move_dictionary)
        if not abs_pos_dict:
            return False
        axis = list(abs_pos_dict.keys())[0]
        return self.move_axis_absolute(axis, abs_pos_dict[axis], wait_until_done)

    def stop(self):
        """Stop all stage movement abruptly."""
        pass

    def switch_mode(self, mode="normal", exposure_times=None, sweep_times=None):
        """Switch Galvo stage working mode.

        Parameters
        ----------
        mode : str
            Name of the stage working mode
            Current supported modes are: confocal-projection, normal
        exposure_times : dict
            Dictionary of exposure times for each channel
        sweep_times : dict
            Dictionary of sweep times for each channel
        """
        self.exposure_times = exposure_times
        self.sweep_times = sweep_times
        if mode == "normal":
            if self.ao_task is None:
                self.ao_task = nidaqmx.Task()
                self.ao_task.ao_channels.add_ao_voltage_chan(self.axes_channels[0])
            self.move_axis_absolute(self.axes[0], float(self.configuration["experiment"]["StageParameters"][self.axes[0]]))
        elif self.ao_task:
            self.ao_task.stop()
            self.ao_task.close()
            self.ao_task = None


# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging

# Third Party Imports
import nidaqmx
from nidaqmx.constants import AcquisitionType

# Local Imports
from navigate.model.devices.galvo.galvo_base import GalvoBase

# # Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class GalvoNI(GalvoBase):
    """GalvoNI Class - NI DAQ Control of Galvanometers"""

    def __init__(self, microscope_name, device_connection, configuration, galvo_id=0):
        """Initialize the GalvoNI class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : dict
            Dictionary of device connections.
        configuration : dict
            Dictionary of configuration parameters.
        galvo_id : int
            Galvo ID.
        """
        super().__init__(microscope_name, device_connection, configuration, galvo_id)

        #: str: Name of the NI port for galvo control.
        self.trigger_source = configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["trigger_source"]

        #: obj: NI DAQ device connection.
        self.daq = device_connection

    def adjust(self, exposure_times, sweep_times):
        """Adjust the galvo to the readout time

        Parameters
        ----------
        exposure_times : dict
            Dictionary of exposure times for each selected channel
        sweep_times : dict
            Dictionary of sweep times for each selected channel

        Returns
        -------
        waveform_dict
            Dictionary with the adjusted waveforms.
        """
        waveform_dict = super().adjust(exposure_times, sweep_times)

        self.daq.analog_outputs[self.device_config["hardware"]["channel"]] = {
            "trigger_source": self.trigger_source,
            "waveform": waveform_dict,
        }
        return waveform_dict

    def turn_off(self):
        """Turn off the galvo.
        Turns off the galvo. NOTE: This will only work if there isn't another task
        bound to this channel. This should only be called in microscope.terminate().
        """
        try:
            task = nidaqmx.Task()
            task.ao_channels.add_ao_voltage_chan(
                self.device_config["hardware"]["channel"]
            )
            task.write([0], auto_start=True)
            task.stop()
            task.close()
        except Exception as e:
            print(f"Galvo turn_off error: {e}")



import unittest
import pytest

try:
    from pyvcam import pvc
    from pyvcam import camera
    from pyvcam import constants as const
except ModuleNotFoundError:
    print("Install pyvcam.")


class CameraConstructionTests(unittest.TestCase):
    def setUp(self):
        pvc.init_pvcam()
        try:
            self.test_cam = next(camera.Camera.detect_camera())
        except Exception:
            raise unittest.SkipTest("No available camera found")

    def tearDown(self):
        # if self.test_cam.is_open():
        #     self.test_cam.close()
        pvc.uninit_pvcam()

    @pytest.mark.hardware
    def test_init(self):
        test_cam_name = "test"
        test_cam = camera.Camera(test_cam_name)
        self.assertEqual(test_cam.name, test_cam_name)
        self.assertEqual(test_cam.handle, -1)
        self.assertEqual(test_cam._Camera__is_open, False)

    @pytest.mark.hardware
    def test_get_dd_version(self):
        self.test_cam.open()
        dd_ver = pvc.get_param(
            self.test_cam.handle, const.PARAM_DD_VERSION, const.ATTR_CURRENT
        )
        dd_ver = "{}.{}.{}".format(
            dd_ver & 0xFF00 >> 8, dd_ver & 0x00F0 >> 4, dd_ver & 0x000F
        )
        self.assertEqual(dd_ver, self.test_cam.driver_version)

    @pytest.mark.hardware
    def test_get_dd_version_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.driver_version

    @pytest.mark.hardware
    def test_get_bin_x(self):
        self.test_cam.open()
        try:
            self.assertEqual(self.test_cam.bin_x, 1)  # Defaults to 1
        except AttributeError:
            self.skipTest("test_get_bin_x: This camera does not " "support binning.")

    @pytest.mark.hardware
    def test_get_bin_y(self):
        self.test_cam.open()
        try:
            self.assertEqual(self.test_cam.bin_y, 1)  # Defaults to 1
        except AttributeError:
            self.skipTest("test_get_bin_y: This camera does not " "support binning.")

    @pytest.mark.hardware
    def test_get_chip_name(self):
        self.test_cam.open()
        chip_name = pvc.get_param(
            self.test_cam.handle, const.PARAM_CHIP_NAME, const.ATTR_CURRENT
        )
        self.assertEqual(chip_name, self.test_cam.chip_name)

    @pytest.mark.hardware
    def test_get_chip_name_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.chip_name

    @pytest.mark.hardware
    def test_get_serial_no(self):
        self.test_cam.open()
        ser_no = pvc.get_param(
            self.test_cam.handle, const.PARAM_HEAD_SER_NUM_ALPHA, const.ATTR_CURRENT
        )
        self.assertEqual(ser_no, self.test_cam.serial_no)

    @pytest.mark.hardware
    def test_get_speed_table_index(self):
        self.test_cam.open()
        spdtab_index = pvc.get_param(
            self.test_cam.handle, const.PARAM_SPDTAB_INDEX, const.ATTR_CURRENT
        )
        self.assertEqual(spdtab_index, self.test_cam.speed_table_index)

    @pytest.mark.hardware
    def test_get_speed_table_index_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.speed_table_index

    @pytest.mark.hardware
    def test_set_speed_table_index(self):
        self.test_cam.open()
        num_entries = self.test_cam.get_param(
            const.PARAM_SPDTAB_INDEX, const.ATTR_COUNT
        )
        for i in range(num_entries):
            self.test_cam.speed_table_index = i
            self.assertEqual(i, self.test_cam.speed_table_index)

    @pytest.mark.hardware
    def test_set_speed_table_index_out_of_bounds(self):
        self.test_cam.open()
        num_entries = self.test_cam.get_param(
            const.PARAM_SPDTAB_INDEX, const.ATTR_COUNT
        )
        with self.assertRaises(ValueError):
            self.test_cam.speed_table_index = num_entries

    @pytest.mark.hardware
    def test_set_speed_table_index_no_open_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.speed_table_index = 0

    @pytest.mark.hardware
    def test_get_readout_port(self):
        self.test_cam.open()
        readout_port = pvc.get_param(
            self.test_cam.handle, const.PARAM_READOUT_PORT, const.ATTR_CURRENT
        )
        self.assertEqual(readout_port, self.test_cam.readout_port)

    @pytest.mark.hardware
    def test_get_readout_port_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.readout_port

    @pytest.mark.hardware
    def test_set_readout_port_index(self):
        self.test_cam.open()
        num_entries = self.test_cam.get_param(
            const.PARAM_READOUT_PORT, const.ATTR_COUNT
        )
        for i in range(num_entries):
            self.test_cam.readout_port = i
            self.assertEqual(i, self.test_cam.readout_port)

    @pytest.mark.hardware
    def test_set_readout_port_out_of_bounds(self):
        self.test_cam.open()
        num_entries = self.test_cam.get_param(
            const.PARAM_READOUT_PORT, const.ATTR_COUNT
        )
        with self.assertRaises(ValueError):
            self.test_cam.readout_port = num_entries

    @pytest.mark.hardware
    def test_set_readout_port_no_open_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.readout_port = 0

    @pytest.mark.hardware
    def test_get_bit_depth(self):
        self.test_cam.open()
        curr_bit_depth = pvc.get_param(
            self.test_cam.handle, const.PARAM_BIT_DEPTH, const.ATTR_CURRENT
        )
        self.assertEqual(curr_bit_depth, self.test_cam.bit_depth)

    @pytest.mark.hardware
    def test_get_bit_depth_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.bit_depth

    @pytest.mark.hardware
    def test_get_pix_time(self):
        self.test_cam.open()
        curr_pix_time = pvc.get_param(
            self.test_cam.handle, const.PARAM_PIX_TIME, const.ATTR_CURRENT
        )
        self.assertEqual(curr_pix_time, self.test_cam.pix_time)

    @pytest.mark.hardware
    def test_get_pix_time_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.pix_time

    @pytest.mark.hardware
    def test_get_gain(self):
        self.test_cam.open()
        curr_gain_index = pvc.get_param(
            self.test_cam.handle, const.PARAM_GAIN_INDEX, const.ATTR_CURRENT
        )
        self.assertEqual(curr_gain_index, self.test_cam.gain)

    @pytest.mark.hardware
    def test_get_gain_fail(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.gain

    @pytest.mark.hardware
    def test_set_gain(self):
        self.test_cam.open()
        max_gain = pvc.get_param(
            self.test_cam.handle, const.PARAM_GAIN_INDEX, const.ATTR_MAX
        )
        for i in range(1, max_gain + 1):
            self.test_cam.gain = i
            curr_gain_index = pvc.get_param(
                self.test_cam.handle, const.PARAM_GAIN_INDEX, const.ATTR_CURRENT
            )
            self.assertEqual(curr_gain_index, self.test_cam.gain)

    @pytest.mark.hardware
    def test_set_gain_less_than_zero(self):
        self.test_cam.open()
        with self.assertRaises(ValueError):
            self.test_cam.gain = -1

    @pytest.mark.hardware
    def test_set_gain_more_than_max(self):
        self.test_cam.open()
        max_gain = pvc.get_param(
            self.test_cam.handle, const.PARAM_GAIN_INDEX, const.ATTR_MAX
        )
        with self.assertRaises(ValueError):
            self.test_cam.gain = max_gain + 1

    @pytest.mark.hardware
    def test_set_gain_no_open(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.gain = 1

    @pytest.mark.hardware
    def test_adc_offset(self):
        self.test_cam.open()
        curr_adc_offset = pvc.get_param(
            self.test_cam.handle, const.PARAM_ADC_OFFSET, const.ATTR_CURRENT
        )
        self.assertEqual(curr_adc_offset, self.test_cam.adc_offset)

    @pytest.mark.hardware
    def test_adc_offset_no_open(self):
        # self.assertRaises(RuntimeError, getattr, self.test_cam, "adc_offset")
        with self.assertRaises(RuntimeError):
            self.test_cam.adc_offset

    @pytest.mark.hardware
    def test_get_clear_mode(self):
        self.test_cam.open()
        curr_clear_mode = pvc.get_param(
            self.test_cam.handle, const.PARAM_CLEAR_MODE, const.ATTR_CURRENT
        )
        self.assertEqual(curr_clear_mode, self.test_cam.clear_mode)
        # reversed_dict = {val: key for key, val in const.clear_modes.items()}
        # self.assertEqual(reversed_dict[curr_clear_mode],
        #                  self.test_cam.clear_mode)

    @pytest.mark.hardware
    def test_get_clear_mode_no_open(self):
        with self.assertRaises(RuntimeError):
            self.test_cam.clear_mode

    @pytest.mark.hardware
    def test_set_clear_mode_by_name(self):
        self.test_cam.open()
        for mode in self.test_cam.clear_modes.values():
            self.test_cam.clear_mode = mode
            curr_clear_mode = pvc.get_param(
                self.test_cam.handle, const.PARAM_CLEAR_MODE, const.ATTR_CURRENT
            )
            self.assertEqual(curr_clear_mode, self.test_cam.clear_mode)

    @pytest.mark.hardware
    def test_set_clear_mode_by_value(self):
        self.test_cam.open()
        for mode in self.test_cam.clear_modes.values():
            self.test_cam.clear_mode = mode
            curr_clear_mode = pvc.get_param(
                self.test_cam.handle, const.PARAM_CLEAR_MODE, const.ATTR_CURRENT
            )
            self.assertEqual(curr_clear_mode, self.test_cam.clear_mode)

    @pytest.mark.hardware
    def test_set_clear_mode_bad_name_fail(self):
        self.test_cam.open()
        self.assertRaises(ValueError, setattr, self.test_cam, "clear_mode", "")

    @pytest.mark.hardware
    def test_set_clear_mode_bad_value_fail(self):
        self.test_cam.open()
        self.assertRaises(ValueError, setattr, self.test_cam, "clear_mode", -1)

    @pytest.mark.hardware
    def test_set_clear_mode_no_open_fail(self):
        self.assertRaises(AttributeError, setattr, self.test_cam, "clear_mode", 0)

    # TODO: All setters should raise Runtime if not open
    @pytest.mark.hardware
    def test_set_clear_mode_no_open_bad_value_fail(self):
        self.assertRaises(AttributeError, setattr, self.test_cam, "clear_mode", -1)

    @pytest.mark.hardware
    def test_get_temp(self):
        self.test_cam.open()
        curr_temp = pvc.get_param(
            self.test_cam.handle, const.PARAM_TEMP, const.ATTR_CURRENT
        )
        # Less than +/-2% variation of temperature between calls.
        self.assertGreaterEqual(abs(curr_temp / self.test_cam.temp), 0.98)

    @pytest.mark.hardware
    def test_get_temp_no_open_fail(self):
        self.assertRaises(RuntimeError, getattr, self.test_cam, "temp")

    @pytest.mark.hardware
    def test_get_temp_setpoint(self):
        self.test_cam.open()
        curr_temp_setpoint = pvc.get_param(
            self.test_cam.handle, const.PARAM_TEMP_SETPOINT, const.ATTR_CURRENT
        )
        self.assertEqual(curr_temp_setpoint / 100, self.test_cam.temp_setpoint)

    @pytest.mark.hardware
    def test_get_temp_setpoint_no_open_fail(self):
        self.assertRaises(RuntimeError, getattr, self.test_cam, "temp_setpoint")

    @pytest.mark.hardware
    def test_set_temp_setpoint(self):
        self.test_cam.open()
        min_temp = self.test_cam.get_param(const.PARAM_TEMP_SETPOINT, const.ATTR_MIN)
        max_temp = self.test_cam.get_param(const.PARAM_TEMP_SETPOINT, const.ATTR_MAX)
        for i in range(max_temp, min_temp - 1, -1):
            self.test_cam.temp_setpoint = i
            self.assertEqual(i / 100, self.test_cam.temp_setpoint)

    @pytest.mark.hardware
    def test_set_temp_to_high_fail(self):
        self.test_cam.open()
        one_over_max_temp = (
            self.test_cam.get_param(const.PARAM_TEMP_SETPOINT, const.ATTR_MAX) + 1
        )
        self.assertRaises(
            ValueError, setattr, self.test_cam, "temp_setpoint", one_over_max_temp
        )

    @pytest.mark.hardware
    def test_set_temp_to_low_fail(self):
        self.test_cam.open()
        one_below_max_temp = (
            self.test_cam.get_param(const.PARAM_TEMP_SETPOINT, const.ATTR_MIN) - 1
        )
        self.assertRaises(
            ValueError, setattr, self.test_cam, "temp_setpoint", one_below_max_temp
        )

    @pytest.mark.hardware
    def test_set_temp_no_open_fail(self):
        self.assertRaises(RuntimeError, setattr, self.test_cam, "temp_setpoint", 0)

    @pytest.mark.hardware
    def test_get_exp_res(self):
        self.test_cam.open()
        curr_exp_res_index = pvc.get_param(
            self.test_cam.handle, const.PARAM_EXP_RES_INDEX, const.ATTR_CURRENT
        )
        self.assertEqual(curr_exp_res_index, self.test_cam.exp_res_index)

    @pytest.mark.hardware
    def test_set_exp_res_by_name(self):
        self.test_cam.open()
        for i in range(self.test_cam.exp_res_index):
            self.test_cam.exp_res = self.test_cam.exp_resolutions[i]
            curr_exp_res = pvc.get_param(
                self.test_cam.handle, const.PARAM_EXP_RES, const.ATTR_CURRENT
            )
            self.assertEqual(curr_exp_res, self.test_cam.exp_res)

    @pytest.mark.hardware
    def test_set_exp_res_bad_name_fail(self):
        self.test_cam.open()
        self.assertRaises(ValueError, setattr, self.test_cam, "exp_res", "")

    @pytest.mark.hardware
    def test_set_exp_res_bad_value_fail(self):
        self.test_cam.open()
        self.assertRaises(ValueError, setattr, self.test_cam, "exp_res", -1)

    @pytest.mark.hardware
    def test_set_exp_res_no_open_fail(self):
        self.assertRaises(AttributeError, setattr, self.test_cam, "exp_res", 0)

    @pytest.mark.hardware
    def test_get_sensor_size(self):
        self.test_cam.open()
        rows = pvc.get_param(
            self.test_cam.handle, const.PARAM_SER_SIZE, const.ATTR_CURRENT
        )
        cols = pvc.get_param(
            self.test_cam.handle, const.PARAM_PAR_SIZE, const.ATTR_CURRENT
        )
        self.assertEqual((rows, cols), self.test_cam.sensor_size)

    @pytest.mark.hardware
    def test_get_sensor_no_open_fail(self):
        self.assertRaises(RuntimeError, getattr, self.test_cam, "sensor_size")

    @pytest.mark.hardware
    def test_get_exp_mode(self):
        self.test_cam.open()
        curr_exp_mode = pvc.get_param(
            self.test_cam.handle, const.PARAM_EXPOSURE_MODE, const.ATTR_CURRENT
        )
        self.assertEqual(curr_exp_mode, self.test_cam.exp_mode)

    @pytest.mark.hardware
    def test_get_exp_mode_no_open(self):
        self.assertRaises(RuntimeError, getattr, self.test_cam, "exp_mode")


def main():
    unittest.main()


if __name__ == "__main__":
    main()



from enum import IntEnum

# Enumerated type defs (cEnum.h):


class E_MODAL_T(IntEnum):

    E_MODALCOEF_LEGENDRE = 0
    E_MODALCOEF_ZERNIKE = 1


class E_ZERNIKE_NORM_T(IntEnum):

    E_ZERNIKE_NORM_STD = 0
    E_ZERNIKE_NORM_RMS = 1


class E_PUPIL_DETECTION_T(IntEnum):

    E_PUPIL_FIXED_RADIUS = 0
    E_PUPIL_FIXED_CENTER = 1
    E_PUPIL_AUTOMATIC = 2


class E_PUPIL_COVERING_T(IntEnum):

    E_PUPIL_INSCRIBED = 0
    E_PUPIL_CIRCUMSCRIBED = 1



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from time import time


def function_timer(func):
    """Decorator for evaluating the duration of time necessary to execute a statement.

    Parameters
    ----------
    func : function
        The function to be timed.

    Returns
    -------
    wrap_func : function
        The wrapped function.
    """

    def wrap_func(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        print(f"Function {func.__name__!r} executed in {(t2 - t1):.4f}s")
        return result

    return wrap_func


class FeatureList(object):
    def __init__(self, func):
        self._feature_list = func
        temp = func.__name__
        self.feature_list_name = str.title(temp.replace("_", " "))

    def __call__(self, *args, **kwargs):
        return self._feature_list()


class AcquisitionMode(object):
    def __init__(self, obj):
        self.__obj_class = obj
        self.__is_acquisition_mode = True

    def __call__(self, *args):
        return self.__obj_class(*args)


import time
import numpy as np

from pyvcam import pvc
from pyvcam.camera import Camera

def main():
    pvc.init_pvcam()
    cam = next(Camera.detect_camera())
    cam.open()

    cnt = 0

    # Check status from sequence collect
    cam.start_seq(exp_time=1000, num_frames=2)

    acquisitionInProgress = True
    while acquisitionInProgress:

        frameStatus = cam.check_frame_status()
        print('Seq frame status: ' + frameStatus)

        if frameStatus == 'READOUT_NOT_ACTIVE' or frameStatus == 'FRAME_AVAILABLE' or frameStatus == 'READOUT_COMPLETE' or frameStatus == 'READOUT_FAILED':
            acquisitionInProgress = False

        if acquisitionInProgress:
            time.sleep(0.05)

    if frameStatus != 'READOUT_FAILED':
        frame, fps, frame_count = cam.poll_frame()

        low = np.amin(frame['pixel_data'])
        high = np.amax(frame['pixel_data'])
        average = np.average(frame['pixel_data'])

        print('Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Rate: {:.1f}\tFrame Count: {:.0f}'.format(low, high, average, fps, frame_count))

    cam.finish()

    frameStatus = cam.check_frame_status()
    print('Seq post-acquisition frame status: ' + frameStatus + '\n')

    # Check status from live collect. Status will only report FRAME_AVAILABLE between acquisitions, so an indeterminate number of frames are needed
    # before we catch the FRAME_AVAILABLE status
    cam.start_live(exp_time=10)

    acquisitionInProgress = True
    while acquisitionInProgress:

        frameStatus = cam.check_frame_status()
        print('Live frame status: ' + frameStatus)

        if frameStatus == 'READOUT_NOT_ACTIVE' or frameStatus == 'FRAME_AVAILABLE' or frameStatus == 'READOUT_FAILED':
            acquisitionInProgress = False

        if acquisitionInProgress:
            time.sleep(0.05)

    if frameStatus != 'READOUT_FAILED':
        frame, fps, frame_count = cam.poll_frame()

        low = np.amin(frame['pixel_data'])
        high = np.amax(frame['pixel_data'])
        average = np.average(frame['pixel_data'])

        print('Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Rate: {:.1f}\tFrame Count: {:.0f}'.format(low, high, average, fps, frame_count))

    cam.finish()

    frameStatus = cam.check_frame_status()
    print('Live post-acquisition frame status: ' + frameStatus)

    cam.close()

    pvc.uninit_pvcam()

    # print('Total frames: {}\n'.format(cnt))


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
import os
import subprocess

# Third Party Imports

# Local Imports


def get_git_revision_hash() -> str:
    """Get the git commit hash of the current repository.

    This function will return the git commit hash of the current repository, or None if
    the current directory is not a git repository.

    Returns:
    --------
    commit_hash : str
        The git commit hash of the current repository, or None if the current directory
        is not a git repository
    """
    working_directory = os.getcwd()
    file_directory = os.path.abspath(os.path.dirname(__file__))
    os.chdir(file_directory)

    try:
        is_git_repo = (
            subprocess.check_output(
                ["git", "rev-parse", "--is-inside-work-tree"], stderr=subprocess.DEVNULL
            )
            .decode("ascii")
            .strip()
        )
        if is_git_repo:
            commit_hash = (
                subprocess.check_output(["git", "rev-parse", "HEAD"])
                .decode("ascii")
                .strip()
            )
        else:
            commit_hash = None
    except subprocess.CalledProcessError:
        commit_hash = None

    os.chdir(working_directory)
    return commit_hash


def get_version_from_file(file_name="VERSION"):
    """Retrieve the version from a file in the same directory as this file.

    The VERSION file located in src/navigate/VERSION contains the version of the
    package when deployed to PyPI.

    Parameters
    ----------
    file_name : str
        The name of the file to read the version from. Default is 'VERSION'

    Returns
    -------
    version : str
        The version of the package
    """
    absolute_path = os.path.abspath(__file__)
    file_path = os.path.join(os.path.dirname(absolute_path), file_name)

    try:
        with open(file_path, "r") as file:
            version = file.read().strip()
        return version
    except FileNotFoundError:
        return "unknown"


__commit__ = get_git_revision_hash()
if __commit__ is None:
    __commit__ = get_version_from_file()



import numpy as np
import time

from pyvcam import pvc
from pyvcam import constants as const
from pyvcam.camera import Camera

NUM_FRAMES = 200
FRAME_DATA_PATH = r'data.bin'
BUFFER_FRAME_COUNT = 16
WIDTH = 1000
HEIGHT = 1000

# Warning. This test can only succeed if the camera supports meta_data since that is needed to confirm frames were not dropped

def main():
    pvc.init_pvcam()
    cam = next(Camera.detect_camera())
    cam.open()
    cam.meta_data_enabled = True
    cam.set_roi(0, 0, WIDTH, HEIGHT)
    cam.start_live(exp_time=100, buffer_frame_count=BUFFER_FRAME_COUNT, stream_to_disk_path=FRAME_DATA_PATH)

    # Data is streamed to disk in a C++ call-back function invoked directly by PVCAM. To not overburden the system,
    # only poll for frames in python at a slow rate, then exit when the frame count indicates all frames have been
    # written to disk
    while True:
        frame, fps, frame_count = cam.poll_frame()

        if frame_count >= NUM_FRAMES:
            low = np.amin(frame['pixel_data'])
            high = np.amax(frame['pixel_data'])
            average = np.average(frame['pixel_data'])
            print('Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Count:{:.0f} Frame Rate: {:.1f}'.format(low, high, average, frame_count, fps))
            break

        time.sleep(1)

    cam.finish()

    imageFormat = cam.get_param(const.PARAM_IMAGE_FORMAT)
    if imageFormat == const.PL_IMAGE_FORMAT_MONO8:
        BYTES_PER_PIXEL = 1
    else:
        BYTES_PER_PIXEL = 2

    cam.close()
    pvc.uninit_pvcam()

    # Read out meta data from stored frames
    FRAME_ALIGNMENT = 0                   # Typically 0. 32 for Kinetix PCIe
    FRAME_BUFFER_ALIGNMENT = 4096
    META_DATA_FRAME_HEADER_SIZE = 48
    META_DATA_ROI_SIZE = 32
    META_DATA_SIZE = META_DATA_FRAME_HEADER_SIZE + META_DATA_ROI_SIZE
    pixelsPerFrame = WIDTH * HEIGHT
    bytesPerFrameUnpadded = pixelsPerFrame * BYTES_PER_PIXEL + META_DATA_SIZE
    framePad = 0 if FRAME_ALIGNMENT == 0 else FRAME_ALIGNMENT - (bytesPerFrameUnpadded % FRAME_ALIGNMENT)
    bytesPerFrame = bytesPerFrameUnpadded + framePad
    frameBufferAlignmentPad = FRAME_BUFFER_ALIGNMENT - ((BUFFER_FRAME_COUNT * bytesPerFrame) % FRAME_BUFFER_ALIGNMENT)

    with open(FRAME_DATA_PATH, "rb") as f:

        badMetaDataCount = 0
        for frame_index in range(NUM_FRAMES):
            frame_number = frame_index + 1

            # Read frame number from meta data header
            # Every time the circular buffer wraps around, bytes are padded into the file. This is a result of needing to
            # write data in chunks that are multiples of the alignment boundary.
            frameBufferPad = int(frame_index / BUFFER_FRAME_COUNT) * frameBufferAlignmentPad

            FRAME_NUMBER_OFFSET = 5
            filePos = frame_index * bytesPerFrame + FRAME_NUMBER_OFFSET + frameBufferPad
            f.seek(filePos, 0)
            frameNumberBytes = f.read(4)

            frame_number_meta_data = int.from_bytes(frameNumberBytes, 'little')

            if frame_number != frame_number_meta_data:
                badMetaDataCount += 1
                print('Expected: ' + str(frame_number) + ' Observed: ' + str(frame_number_meta_data))

    print('Verified ' + str(NUM_FRAMES) + ' frames:')
    print('  Meta data errors: ' + str(badMetaDataCount))

    if badMetaDataCount > 0:
        print('\nMeta data error troubleshooting:')
        print('  1. Verify FRAME_ALIGNMENT for camera and interface')
        print('  2. Increase exposure time')

if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.
#
# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import logging
import time

# Third Party Imports

# Local Imports
from navigate.model.devices.stages.stage_base import StageBase
from navigate.model.devices.APIs.asi.asi_tiger_controller import (
    TigerController,
    TigerException,
)

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def build_ASI_Stage_connection(com_port, baud_rate=115200):
    """Connect to the ASI Stage

    Parameters
    ----------
    com_port : str
        Communication port for ASI Tiger Controller - e.g., COM1
    baud_rate : int
        Baud rate for ASI Tiger Controller - e.g., 9600

    Returns
    -------
    asi_stage : object
        Successfully initialized stage object.
    """

    # wait until ASI device is ready
    asi_stage = TigerController(com_port, baud_rate)
    asi_stage.connect_to_serial()
    if not asi_stage.is_open():
        raise Exception("ASI stage connection failed.")

    return asi_stage


class ASIStage(StageBase):
    """Applied Scientific Instrumentation (ASI) Stage Class

    ASI Documentation: http://asiimaging.com/docs/products/serial_commands

    ASI Quick Start Guide: http://asiimaging.com/docs/command_quick_start

    Note
    ----
    ASI firmware requires all distances to be in a 10th of a micron.

    Warning
    -------
        Do not ever change the F axis. This will alter the relative position of each
        FTP stilt, adding strain to the system. Only move the Z axis, which will
        change both stilt positions simultaneously.
    """

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the ASI Stage connection.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocessing.managers.DictProxy
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration, device_id)

        # Default axes mapping
        axes_mapping = {"x": "Z", "y": "Y", "z": "X", "f": "M"}
        if not self.axes_mapping:
            self.axes_mapping = {
                axis: axes_mapping[axis] for axis in self.axes if axis in axes_mapping
            }
        #: Mapping of axes to ASI axes
        else:
            # Force cast axes to uppercase
            self.axes_mapping = {k: v.upper() for k, v in self.axes_mapping.items()}

        self.asi_axes = dict(map(lambda v: (v[1], v[0]), self.axes_mapping.items()))

        # Set feedback alignment values - Default to 85 if not specified
        if self.stage_feedback is None:
            feedback_alignment = {axis: 85 for axis in self.asi_axes}
        else:
            feedback_alignment = {
                axis: self.stage_feedback
                for axis, self.stage_feedback in zip(self.asi_axes, self.stage_feedback)
            }

        self.tiger_controller = device_connection
        if device_connection is not None:
            # Set feedback alignment values
            for ax, aa in feedback_alignment.items():
                self.tiger_controller.set_feedback_alignment(ax, aa)
            logger.debug("ASI Stage Feedback Alignment Settings:", feedback_alignment)

            # Set finishing accuracy to half of the minimum pixel size we will use
            # pixel size is in microns, finishing accuracy is in mm
            # TODO: check this over all microscopes sharing this stage,
            #       not just the current one
            finishing_accuracy = (
                0.001
                * min(
                    list(
                        configuration["configuration"]["microscopes"][microscope_name][
                            "zoom"
                        ]["pixel_size"].values()
                    )
                )
                / 2
            )
            # If this is changing, the stage must be power cycled for these changes to
            # take effect.
            for ax in self.asi_axes.keys():
                if self.asi_axes[ax] == "theta":
                    self.tiger_controller.set_finishing_accuracy(ax, 0.003013)
                    self.tiger_controller.set_error(ax, 0.1)
                else:
                    self.tiger_controller.set_finishing_accuracy(ax, finishing_accuracy)
                    self.tiger_controller.set_error(ax, 1.2 * finishing_accuracy)

            # Set backlash to 0 (less accurate)
            for ax in self.asi_axes.keys():
                if self.asi_axes[ax] == "theta":
                    self.tiger_controller.set_backlash(ax, 0.1)
                self.tiger_controller.set_backlash(ax, 0.0)

            # Speed optimizations - Set speed to 90% of maximum on each axis
            self.set_speed(percent=0.9)

    def __del__(self):
        """Delete the ASI Stage connection."""
        try:
            if self.tiger_controller is not None:
                self.tiger_controller.disconnect_from_serial()
                logger.debug("ASI stage connection closed")
        except (AttributeError, BaseException) as e:
            logger.exception("ASI Stage Exception", e)
            raise

    def get_axis_position(self, axis):
        """Get position of specific axis

        Parameters
        ----------
        axis : str
            Axis to get position of

        Returns
        -------
            position: float
        """
        try:
            axis = self.axes_mapping[axis]
            pos = self.tiger_controller.get_axis_position_um(axis)
        except TigerException:
            return float("inf")
        except KeyError as e:
            logger.exception(f"ASI Stage - KeyError in get_axis_position: {e}")
            return float("inf")
        return pos

    def report_position(self):
        """Reports the position for all axes in microns, and create
        position dictionary."""
        try:
            # positions from the device are in microns
            pos_dict = self.tiger_controller.get_position(list(self.asi_axes.keys()))
            for axis, pos in pos_dict.items():
                ax = self.asi_axes[axis]
                if ax == "theta":
                    setattr(self, f"{ax}_pos", float(pos) / 1000.0)
                else:
                    setattr(self, f"{ax}_pos", float(pos) / 10.0)
        except TigerException as e:
            logger.exception("ASI Stage Exception", e)

        return self.get_position_dict()

    def move_axis_absolute(self, axis, abs_pos, wait_until_done=False):
        """Move stage along a single axis.

        Move absolute command for ASI is MOVE [Axis]=[units 1/10 microns]

        Parameters
        ----------
        axis : str
            An axis prefix in move_dictionary. For example, axis='x' corresponds to
            'x_abs', 'x_min', etc.
        abs_pos : float
            Absolute position value
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """
        if axis not in self.axes_mapping:
            return False

        axis_abs = self.get_abs_position(axis, abs_pos)
        if axis_abs == -1e50:
            print("axis abs false")
            return False

        # Move stage
        try:
            if axis == "theta":
                self.tiger_controller.move_axis(
                    self.axes_mapping[axis], axis_abs * 1000
                )
            else:
                # The 10 is to account for the ASI units, 1/10 of a micron
                self.tiger_controller.move_axis(self.axes_mapping[axis], axis_abs * 10)

        except TigerException as e:
            print(
                f"ASI stage move axis absolute failed or is trying to move out of "
                f"range: {e}"
            )
            logger.exception("ASI Stage Exception", e)
            return False

        if wait_until_done:
            self.tiger_controller.wait_for_device()
        return True

    def verify_move(self, move_dictionary):
        """Don't submit a move command for axes that aren't moving.
        The Tiger controller wait time for each axis is additive.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement. Includes 'x_abs', etc. for
            one or more axes. Expect values in micrometers, except for theta, which is
            in degrees.

        Returns
        -------
        res_dict : dict
            A dictionary of values required for movement. Includes 'x_abs', etc. for
            one or more axes. Expect values in micrometers, except for theta, which is
            in degrees.
        """
        res_dict = {}
        for axis, val in move_dictionary.items():
            curr_pos = getattr(self, f"{axis}_pos", None)
            if curr_pos != val:
                res_dict[axis] = val
        return res_dict

    def move_absolute(self, move_dictionary, wait_until_done=False):
        """Move Absolute Method.
        XYZ Values should remain in microns for the ASI API
        Theta Values are not accepted.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement. Includes 'x_abs', etc. for
            one or more axes. Expect values in micrometers, except for theta, which is
            in degrees.
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        success : bool
            Was the move successful?
        """
        abs_pos_dict = self.verify_abs_position(move_dictionary)
        if not abs_pos_dict:
            return False
        abs_pos_dict = self.verify_move(abs_pos_dict)
        if len(abs_pos_dict) == 0:
            return

        # This is to account for the asi 1/10 of a micron units
        pos_dict = {
            self.axes_mapping[axis]: pos * 1000 if axis == "theta" else pos * 10
            for axis, pos in abs_pos_dict.items()
        }
        try:
            self.tiger_controller.move(pos_dict)
        except TigerException as e:
            print(
                f"ASI stage move axis absolute failed or is trying to move out of "
                f"range: {e}"
            )
            logger.exception("ASI Stage Exception", e)
            return False
        if wait_until_done:
            self.tiger_controller.wait_for_device()

        return True

    def stop(self):
        """Stop all stage movement abruptly."""
        try:
            self.tiger_controller.stop()
        except TigerException as e:
            print(f"ASI stage halt command failed: {e}")
            logger.exception("ASI Stage Exception", e)

    def set_speed(self, velocity_dict=None, percent=None):
        """Set scan velocity.

        Parameters
        ----------
        velocity_dict: dict
            velocity for specific axis
            {'x': float, 'y': float, 'z': float}
        percent : float
            Percent of maximum speed

        Returns
        -------
        success: bool
            Was the setting successful?
        """
        if percent is not None:
            try:
                self.tiger_controller.set_speed_as_percent_max(percent)
            except TigerException as e:
                print(f"ASI Controller failed to set speed as a percent: {e}")
                return False
        else:
            try:
                self.tiger_controller.set_speed(velocity_dict)
            except TigerException:
                return False
            except KeyError as e:
                logger.exception(f"ASI Stage - KeyError in set_speed: {e}")
                return False
        return True

    def get_speed(self, axis):
        """Get scan velocity of the axis.

        Parameters
        ----------
        axis: str
            axis name, such as 'x', 'y', 'z'

        Returns
        -------
        velocity: float
            Velocity
        """
        try:
            velocity = self.tiger_controller.get_speed(self.axes_mapping[axis])
        except TigerException:
            return 0
        except KeyError as e:
            logger.exception(f"ASI Stage - KeyError in get_speed: {e}")
            return 0
        return velocity

    def scanr(self, start_position_mm, end_position_mm, enc_divide, axis="z"):
        """Set scan range

        Parameters
        ----------
        start_position_mm: float
            scan start position
        end_position_mm: float
            scan end position
        enc_divide: float
            Step size desired.
        axis: str
            fast axis name

        Returns
        -------
        success: bool
            Was the setting successful?
        """
        try:
            axis = self.axes_mapping[axis]
            self.tiger_controller.scanr(
                start_position_mm, end_position_mm, enc_divide, axis
            )
        except TigerException as e:
            logger.exception(f"TigerException: {e}")
            print(logger.exception())
            return False
        except KeyError as e:
            logger.exception(f"ASI Stage - KeyError in scanr: {e}")
            return False
        # if wait_until_done:
        #     self.tiger_controller.wait_for_device()

        return True
        # return True

    def scanv(
        self, start_position_mm, end_position_mm, number_of_lines, overshoot, axis="z"
    ):
        """Set scan range

        Parameters
        ----------
        start_position_mm: float
            scan start position
        end_position_mm: float
            scan end position
        number of lines: int
            number of steps.
        overshoot: float
            overshoot_time ms
        axis: str
            fast axis name

        Returns
        -------
        success: bool
            Was the setting successful?
        """
        try:
            axis = self.axes_mapping[axis]
            self.tiger_controller.scanv(
                start_position_mm, end_position_mm, number_of_lines, overshoot, axis
            )
        except TigerException as e:
            logger.exception(f"TigerException: {e}")
            print(logger.exception())
            return False
        except KeyError as e:
            logger.exception(f"ASI Stage - KeyError in scanr: {e}")
            return False
        return True

    def start_scan(self, axis):
        """Start scan state machine

        Parameters
        ----------
        axis: str
            fast axis name, such as 'x', 'y', and 'z'

        Returns
        -------
        success: bool
            Was it successful?

        """
        try:
            axis = self.axes_mapping[axis]
            self.tiger_controller.start_scan(axis)
        except TigerException as e:
            logger.exception(f"TigerException: {e}")
            return False
        except KeyError as e:
            logger.exception(f"ASI Stage - KeyError in start_scan: {e}")
            return False
        return True

    def stop_scan(self):
        """Stop scan"""
        try:
            self.tiger_controller.stop_scan()
        except TigerException as e:
            logger.exception("ASI Stage Exception", e)

    def wait_until_complete(self, axis):
        try:
            while self.tiger_controller.is_axis_busy(axis):
                time.sleep(0.1)
        except TigerException as e:
            print(f"ASI Stage Exception {e}")
            logger.exception(f"ASI Stage Exception {e}")
            return False
        return True



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)
"""
see:


"""


class Tooltip:
    """Create a tooltip for a given widget when the mouse hovers over it.

    http://stackoverflow.com/questions/3221956/
       what-is-the-simplest-way-to-make-tooltips-in-tkinter/36221216#36221216

    http://www.daniweb.com/programming/software-development/
           code/484591/a-tooltip-class-for-tkinter

    - Originally written by vegaseat on 2014.09.09.

    - Modified to include a delay time by Victor Zaccardo on 2016.03.25.

    - Modified
        - to correct extreme right and extreme bottom behavior,
        - to stay inside the screen whenever the tooltip might go out on
          the top but still the screen is higher than the tooltip,
        - to use the more flexible mouse positioning,
        - to add customizable background color, padding, waittime and
          wraplength on creation
      by Alberto Vassena on 2016.11.05.
    """

    def __init__(
        self,
        widget,
        *,
        bg="#FFFFEA",
        pad=(5, 3, 5, 3),
        text="widget info",
        waittime=400,
        wraplength=250
    ):
        """Initialize the ToolTip.

        Parameters
        ----------
        widget : tkinter widget
            The widget to which this tooltip is bound.
        bg : str, optional
            Background color for the tooltip. The default is "#FFFFEA".
        pad : tuple, optional
            Padding to be left around the text. The default is (5, 3, 5, 3).
        text : str, optional
            Text to be displayed in the tooltip. The default is "widget info".
        waittime : int, optional
            Delay in miliseconds before the tooltip shows up. The default is 400.
        wraplength : int, optional
            Wrap length for the tooltip text. The default is 250.
        """
        #: int: Delay in miliseconds before the tooltip shows up.
        self.waittime = waittime  # in miliseconds, originally 500
        #: int: Wrap length for the tooltip text.
        self.wraplength = wraplength  # in pixels, originally 180
        #: tk.Widget: The widget to which this tooltip is bound.
        self.widget = widget
        #: str: Text to be displayed in the tooltip.
        self.text = text
        # self.widget.bind("<Button-1>", self.onEnter)
        self.widget.bind("<Enter>", self.onEnter)
        self.widget.bind("<Leave>", self.onLeave)
        self.widget.bind("<ButtonPress>", self.onLeave)
        #: str: Background color for the tooltip.
        self.bg = bg
        #: tuple: Padding to be left around the text.
        self.pad = pad
        #: int: The id of the scheduled event.
        self.id = None
        #: tk.Toplevel: The tooltip toplevel.
        self.tw = None

    def onEnter(self, event=None):
        """Called when the mouse enters the widget.

        Parameters
        ----------
        event : tk.Event, optional
            The event that called this function. The default is None.
        """
        self.schedule()

    def onLeave(self, event=None):
        """Called when the mouse leaves the widget.

        Parameters
        ----------
        event : tk.Event, optional
            The event that called this function. The default is None.
        """
        self.unschedule()
        self.hide()

    def schedule(self):
        """Schedule the event."""
        self.unschedule()
        self.id = self.widget.after(self.waittime, self.show)

    def unschedule(self):
        """Unschedule the event."""
        id_ = self.id
        self.id = None
        if id_:
            self.widget.after_cancel(id_)

    def show(self):
        """Display the tooltip."""

        def tip_pos_calculator(widget, label, *, tip_delta=(10, 5), pad=(5, 3, 5, 3)):
            """Calculate the position of the tooltip on the screen.

            Parameters
            ----------
            widget : tk.Widget
                The widget to which this tooltip is bound.
            label : tk.Label
                The label containing the tooltip text.
            tip_delta : tuple, optional
                The offset of the tooltip from the mouse pointer.
                The default is (10, 5).
            pad : tuple, optional
                Padding to be left around the text. The default is (5, 3, 5, 3).

            Returns
            -------
            tuple
                The position of the tooltip on the screen.
            """

            w = widget

            s_width, s_height = w.winfo_screenwidth(), w.winfo_screenheight()

            width, height = (
                pad[0] + label.winfo_reqwidth() + pad[2],
                pad[1] + label.winfo_reqheight() + pad[3],
            )

            mouse_x, mouse_y = w.winfo_pointerxy()

            x1, y1 = mouse_x + tip_delta[0], mouse_y + tip_delta[1]
            x2, y2 = x1 + width, y1 + height

            x_delta = x2 - s_width
            if x_delta < 0:
                x_delta = 0
            y_delta = y2 - s_height
            if y_delta < 0:
                y_delta = 0

            offscreen = (x_delta, y_delta) != (0, 0)

            if offscreen:

                if x_delta:
                    x1 = mouse_x - tip_delta[0] - width

                if y_delta:
                    y1 = mouse_y - tip_delta[1] - height

            offscreen_again = y1 < 0  # out on the top

            if offscreen_again:
                # No further checks will be done.

                # TIP:
                # A further mod might automagically augment the
                # wraplength when the tooltip is too high to be
                # kept inside the screen.
                y1 = 0

            return x1, y1

        bg = self.bg
        pad = self.pad
        widget = self.widget

        # creates a toplevel window
        self.tw = tk.Toplevel(widget)

        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)

        win = tk.Frame(self.tw, background=bg, borderwidth=0)
        label = tk.Label(
            win,
            text=self.text,
            justify=tk.LEFT,
            background=bg,
            relief=tk.SOLID,
            borderwidth=0,
            wraplength=self.wraplength,
        )

        label.grid(padx=(pad[0], pad[2]), pady=(pad[1], pad[3]), sticky=tk.NSEW)
        win.grid()

        x, y = tip_pos_calculator(widget, label)

        self.tw.wm_geometry("+%d+%d" % (x, y))

    def hide(self):
        """Hide the tooltip."""
        tw = self.tw
        if tw:
            tw.destroy()
        self.tw = None



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Imports
import logging
from threading import Lock
import traceback

# Third Party Imports
import nidaqmx
import numpy as np

# Local Imports
from navigate.model.devices.daq.daq_base import DAQBase
from navigate.tools.waveform_template_funcs import get_waveform_template_parameters

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class NIDAQ(DAQBase):
    """NIDAQ class for Control of NI Data Acquisition Cards."""

    def __init__(self, configuration):
        """Initialize NIDAQ class.

        Parameters
        ----------
        configuration : dict
            Configuration dictionary.
        """
        super().__init__(configuration)

        #: obj: NI DAQmx task for camera trigger
        self.camera_trigger_task = None

        #: obj: NI DAQmx task for master trigger
        self.master_trigger_task = None

        #: obj: NI DAQmx task for laser switching
        self.laser_switching_task = None

        #: str: Trigger mode. Self-trigger or external-trigger.
        self.trigger_mode = "self-trigger"  # self-trigger, external-trigger

        #: str: NI DAQmx port for laser switching
        self.external_trigger = None

        # keep track of analog outputs and their waveforms
        #: dict: Analog outputs.
        self.analog_outputs = {}

        #: dict: NI DAQmx tasks for analog output.
        self.analog_output_tasks = {}

        #: float: Number of samples.
        self.n_sample = None

        #: str: Current channel key.
        self.current_channel_key = ""

        #: bool: Flag for updating analog task.
        self.is_updating_analog_task = False

        #: bool: Flag for waiting to run.
        self.wait_to_run_lock = Lock()

    def __del__(self):
        """Destructor."""
        if self.camera_trigger_task is not None:
            self.stop_acquisition()

    def set_external_trigger(self, external_trigger=None):
        """Set trigger mode.

        Parameters
        ----------
        external_trigger : nidaqmx.Task
            Task for external triggering
        """
        self.trigger_mode = (
            "self-trigger" if external_trigger is None else "external-trigger"
        )
        self.external_trigger = external_trigger

        # change trigger mode during acquisition in a feature
        if self.trigger_mode == "self-trigger":
            self.create_master_trigger_task()
            trigger_source = self.configuration["configuration"]["microscopes"][
                self.microscope_name
            ]["daq"]["trigger_source"]

            # set camera task trigger source
            try:
                self.camera_trigger_task.stop()
            except Exception:
                logger.debug(
                    f"DAQ NI - Error switching the camera trigger source: "
                    f"{traceback.format_exc()}"
                )
            self.camera_trigger_task.triggers.start_trigger.cfg_dig_edge_start_trig(
                trigger_source
            )
            self.camera_trigger_task.triggers.start_trigger.retriggerable = False
            # set analog task trigger source
            for board_name in self.analog_output_tasks.keys():
                try:
                    self.analog_output_tasks[board_name].stop()
                except Exception:
                    logger.debug(
                        f"DAQ NI - Error stopping analog output tasks: "
                        f"{traceback.format_exc()}"
                    )
                self.analog_output_tasks[
                    board_name
                ].triggers.start_trigger.cfg_dig_edge_start_trig(trigger_source)
                self.analog_output_tasks[board_name].register_done_event(None)

        else:
            # close master trigger task
            if self.master_trigger_task:
                try:
                    self.master_trigger_task.stop()
                    self.master_trigger_task.close()
                except Exception:
                    logger.debug(
                        f"DAQ NI - Error stopping master trigger task: "
                        f"{traceback.format_exc()}"
                    )
            self.master_trigger_task = None
            # camera task trigger source
            self.camera_trigger_task.triggers.start_trigger.cfg_dig_edge_start_trig(
                self.external_trigger
            )

            # change camera task to so that it can be triggered again.
            self.camera_trigger_task.triggers.start_trigger.retriggerable = False

            # add callback function to analog tasks
            for board_name in self.analog_output_tasks.keys():
                task = self.analog_output_tasks[board_name]
                task.triggers.start_trigger.cfg_dig_edge_start_trig(
                    self.external_trigger
                )
                task.register_done_event(None)
                task.register_done_event(self.restart_analog_task_callback_func(task))

    @staticmethod
    def restart_analog_task_callback_func(task):
        """Restart analog task callback function.

        Parameters
        ----------
        task : nidaqmx.Task
            Task for analog output

        Returns
        -------
        callback_func : function
            Callback function
        """

        def callback_func(task_handle, status, callback_data):
            try:
                logger.info("Analog Tasks Restarted")
                task.stop()
                task.start()
            except Exception:
                logger.debug(
                    f"DAQ NI - Analog task restart failed {traceback.format_exc()}"
                )
            return status

        return callback_func

    def create_camera_task(self, channel_key):
        """Set up the camera trigger task.

        TTL for triggering the camera. TTL is 4 ms in duration.
        Channel that the TTL is delivered from, and its delay (typically ~10 ms), are
        specified in the configuration.yaml file.

        Parameters
        ----------
        channel_key : str
            Channel key for current channel.
        """
        self.camera_trigger_task = nidaqmx.Task()
        camera_trigger_out_line = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["daq"]["camera_trigger_out_line"]
        camera_high_time = 0.004
        exposure_time = self.exposure_times[channel_key]
        sweep_time = self.sweep_times[channel_key]
        camera_low_time = sweep_time - camera_high_time
        camera_delay = (self.camera_delay_percent / 100) * (exposure_time / 1000)

        self.camera_trigger_task.co_channels.add_co_pulse_chan_time(
            camera_trigger_out_line,
            high_time=camera_high_time,
            low_time=camera_low_time,
            initial_delay=camera_delay,
        )

        # apply waveform templates
        camera_waveform_repeat_num = self.waveform_repeat_num * self.waveform_expand_num
        self.camera_trigger_task.timing.cfg_implicit_timing(
            sample_mode=nidaqmx.constants.AcquisitionType.FINITE,
            samps_per_chan=camera_waveform_repeat_num,
        )

    def create_master_trigger_task(self):
        """Set up the DO master trigger task."""
        self.master_trigger_task = nidaqmx.Task()
        master_trigger_out_line = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["daq"]["master_trigger_out_line"]
        self.master_trigger_task.do_channels.add_do_chan(
            master_trigger_out_line,
            line_grouping=nidaqmx.constants.LineGrouping.CHAN_FOR_ALL_LINES,
        )

    def create_analog_output_tasks(self, channel_key):
        """Create analog output tasks for each board.

        Create a single analog output task for all channels per board. Most NI DAQ cards
        have only one clock for analog output sample timing, and as such all channels
        must be grouped here.

        Parameters
        ----------
        channel_key : str
            Channel key for analog output.
        """
        self.n_sample = int(self.sample_rate * self.sweep_times[channel_key])
        max_sample = self.n_sample * self.waveform_expand_num
        # TODO: GalvoStage and remote_focus waveform are not calculated based on a
        #  same sweep time. There needs some fix.

        # Create one analog output task per board, grouping the channels
        boards = list(set([x.split("/")[0] for x in self.analog_outputs.keys()]))
        for board in boards:
            channel = ", ".join(
                list(
                    [x for x in self.analog_outputs.keys() if x.split("/")[0] == board]
                )
            )
            self.analog_output_tasks[board] = nidaqmx.Task()
            self.analog_output_tasks[board].ao_channels.add_ao_voltage_chan(channel)

            # apply templates to analog tasks
            self.analog_output_tasks[board].timing.cfg_samp_clk_timing(
                rate=self.sample_rate,
                sample_mode=nidaqmx.constants.AcquisitionType.FINITE,
                samps_per_chan=max_sample * self.waveform_repeat_num,
            )

            # triggers = list(
            #     set([v["trigger_source"] for v in self.analog_outputs.values()])
            # )
            # if len(triggers) > 1:
            #     logger.debug(
            #         "NI DAQ - Different triggers provided for each analog channel."
            #         "Defaulting to the first trigger provided."
            #     )
            # self.analog_output_tasks[board].triggers.start_trigger.cfg_dig_edge_start_trig(
            #     triggers[0]
            # )
            # TODO: may change this later to automatically expand the waveform to the
            #  longest
            for k, v in self.analog_outputs.items():
                if k.split("/")[0] == board and len(v["waveform"][channel_key]) < max_sample:
                    v["waveform"][channel_key] = np.hstack(
                        [v["waveform"][channel_key]] * self.waveform_expand_num
                    )
            # Write values to board
            waveforms = np.vstack(
                [
                    v["waveform"][channel_key][:max_sample]
                    for k, v in self.analog_outputs.items()
                    if k.split("/")[0] == board
                ]
            ).squeeze()
            self.analog_output_tasks[board].write(waveforms)

    def prepare_acquisition(self, channel_key):
        """Prepare the acquisition.

        Creates and configures the DAQ tasks.
        Writes the waveforms to each task.

        Parameters
        ----------
        channel_key : str
            Channel key for current channel.
        """
        waveform_template_name = self.configuration["experiment"]["MicroscopeState"][
            "waveform_template"
        ]
        logger.info(f"Waveform Template Name: {waveform_template_name}")
        (
            self.waveform_repeat_num,
            self.waveform_expand_num,
        ) = get_waveform_template_parameters(
            waveform_template_name,
            self.configuration["waveform_templates"],
            self.configuration["experiment"]["MicroscopeState"],
        )

        logger.info(f"Waveform Expand Num = {self.waveform_expand_num}")
        logger.info(f"Waveform Repeat Num = {self.waveform_repeat_num}")

        self.create_camera_task(channel_key)
        self.create_analog_output_tasks(channel_key)
        self.current_channel_key = channel_key
        self.is_updating_analog_task = False
        if self.wait_to_run_lock.locked():
            self.wait_to_run_lock.release()
        # Specify ports, timing, and triggering
        self.set_external_trigger(self.external_trigger)

    def run_acquisition(self):
        """Run DAQ Acquisition.

        Run the tasks for triggering, analog and counter outputs.
        The master trigger initiates all other tasks via a shared trigger
        For this to work, all analog output and counter tasks have to be started so that
        they are waiting for the trigger signal.
        """
        # wait if writing analog tasks
        if self.is_updating_analog_task:
            self.wait_to_run_lock.acquire()
            self.wait_to_run_lock.release()

        if self.camera_trigger_task.is_task_done():
            self.camera_trigger_task.start()
            for task in self.analog_output_tasks.values():
                task.start()

        if self.trigger_mode == "self-trigger":
            self.master_trigger_task.write(
                [False, True, True, True, False], auto_start=True
            )

        try:
            self.camera_trigger_task.wait_until_done(timeout=10000)
            for task in self.analog_output_tasks.values():
                if self.trigger_mode == "self-trigger":
                    task.wait_until_done()
                task.stop()
        except Exception:
            # when triggered from external triggers, sometimes the camera trigger task
            # is done but not actually done, there will a DAQ WARNING message
            logger.debug(
                f"DAQ NI - Wait until tasks done failed - {traceback.format_exc()}"
            )
            pass
        try:
            self.camera_trigger_task.stop()
            if self.trigger_mode == "self-trigger":
                self.master_trigger_task.stop()
        except nidaqmx.DaqError:
            pass

    def stop_acquisition(self):
        """Stop Acquisition.

        Stop all tasks and close them.
        """
        try:
            self.camera_trigger_task.stop()
            self.camera_trigger_task.close()

            if self.trigger_mode == "self-trigger":
                self.master_trigger_task.stop()
                self.master_trigger_task.close()

            for k, task in self.analog_output_tasks.items():
                task.stop()
                task.close()

        except (AttributeError, nidaqmx.errors.DaqError):
            pass

        if self.wait_to_run_lock.locked():
            self.wait_to_run_lock.release()

        self.analog_output_tasks = {}

    def enable_microscope(self, microscope_name):
        """Enable microscope.

        Parameters
        ----------
        microscope_name : str
            Name of microscope to enable.
        """
        if microscope_name != self.microscope_name:
            self.microscope_name = microscope_name
            self.analog_outputs = {}
            self.analog_output_tasks = {}
        
        self.camera_delay_percent = self.configuration["configuration"]["microscopes"][
            microscope_name
        ]["camera"]["delay_percent"]
        self.sample_rate = self.configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["sample_rate"]

        try:
            switching_port = self.configuration["configuration"]["microscopes"][
                self.microscope_name
            ]["daq"]["laser_port_switcher"]
            switching_on_state = self.configuration["configuration"]["microscopes"][
                self.microscope_name
            ]["daq"]["laser_switch_state"]

            if self.laser_switching_task:
                self.laser_switching_task.close()
            self.laser_switching_task = nidaqmx.Task()
            self.laser_switching_task.do_channels.add_do_chan(
                switching_port,
                line_grouping=nidaqmx.constants.LineGrouping.CHAN_FOR_ALL_LINES,
            )
            self.laser_switching_task.write(switching_on_state, auto_start=True)
        except KeyError:
            pass

    def update_analog_task(self, board_name):
        """Update analog task.

        Parameters
        ----------
        board_name : str
            Name of board to update.

        Returns
        -------
        bool
            True if task is updated, False otherwise.
        """
        # if there is no such analog task,
        # it means it's not acquiring and nothing needs to do.
        if board_name not in self.analog_output_tasks:
            return False
        # can't update an analog task while updating one.
        if self.is_updating_analog_task:
            return False

        self.wait_to_run_lock.acquire()
        self.is_updating_analog_task = True

        try:
            # this function waits only happens when
            # interacting through GUI in continuous mode,

            # updating an analog task happens after the task
            # is done when running a feature, so it will check and return immediately.
            self.analog_output_tasks[board_name].wait_until_done(timeout=1.0)
            self.analog_output_tasks[board_name].stop()

            # Write values to board
            waveforms = np.vstack(
                [
                    v["waveform"][self.current_channel_key][: self.n_sample]
                    for k, v in self.analog_outputs.items()
                    if k.split("/")[0] == board_name
                ]
            ).squeeze()
            self.analog_output_tasks[board_name].write(waveforms)
        except Exception:
            logger.debug(
                f"DAQ NI - Could not update analog task: {traceback.format_exc()}"
            )
            for board in self.analog_output_tasks.keys():
                try:
                    self.analog_output_tasks[board].stop()
                    self.analog_output_tasks[board].close()
                except Exception:
                    logger.debug(
                        f"DAQ NI - Could not stop analog tasks: "
                        f"{traceback.format_exc()}"
                    )

            self.create_analog_output_tasks(self.current_channel_key)

        self.is_updating_analog_task = False
        self.wait_to_run_lock.release()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import logging
import tkinter as tk

# Third party imports

# Local application imports
from navigate.controller.sub_controllers.gui_controller import GUIController


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ChannelSettingController(GUIController):
    """Controller for the channel setting widgets."""

    def __init__(self, view, parent_controller=None, configuration_controller=None):
        """Initialize the ChannelSettingController.

        Parameters
        ----------
        view : navigate.view.channel_setting_view.ChannelSettingView
            The view for the channel setting widgets.
        parent_controller : navigate.controller.main_controller.MainController
            The parent controller.
        configuration_controller : ConfigurationController
            The configuration controller.
        """
        super().__init__(view, parent_controller)

        #: ConfigurationController: The configuration controller.
        self.configuration_controller = configuration_controller

        # num: numbers of channels
        self.num = self.configuration_controller.number_of_channels
        self.view.populate_frame(channels=self.num)

        #: str: The mode of the channel setting controller. Either 'live' or 'stop'.
        self.mode = "stop"

        #: bool: Whether the channel setting controller is in initialization.
        self.in_initialization = True

        #: int: The event id.
        self.event_id = None

        #: dict: The channel setting dictionary.
        self.channel_setting_dict = None

        # widget command binds
        for i in range(self.num):
            channel_vals = self.get_vals_by_channel(i)
            for name in channel_vals:
                channel_vals[name].trace_add("write", self.channel_callback(i, name))

    def set_mode(self, mode="stop"):
        """Set the mode of the channel setting controller.

        Parameters
        ----------
        mode : str
            "stop" or "live"

        Examples
        --------
        >>> self.set_mode("live")
        """

        self.mode = mode
        state = "normal" if mode == "stop" else "disabled"
        state_readonly = "readonly" if mode == "stop" else "disabled"
        for i in range(self.num):
            # State set regardless of operating mode.
            self.view.channel_checks[i].config(state=state)
            self.view.interval_spins[i].config(state="disabled")
            self.view.laser_pulldowns[i]["state"] = state_readonly

            if self.mode != "live" or (
                self.mode == "live" and not self.view.channel_variables[i].get()
            ):
                self.view.exptime_pulldowns[i].config(state=state)

            if not self.view.channel_variables[i].get():
                self.view.laserpower_pulldowns[i].config(state=state)
                self.view.filterwheel_pulldowns[i]["state"] = state_readonly
                self.view.defocus_spins[i].config(state=state)

    def initialize(self):
        """Populates the laser and filter wheel options in the View.

        Examples
        --------
        >>> self.initialize()
        """
        setting_dict = self.configuration_controller.channels_info
        for i in range(self.num):
            self.view.laser_pulldowns[i]["values"] = setting_dict["laser"]
            self.view.filterwheel_pulldowns[i]["values"] = setting_dict["filter"]
        self.show_verbose_info("channel has been initialized")

    def populate_experiment_values(self, setting_dict):
        """Populates the View with the values from the setting dictionary.

        Set channel values according to channel id
        the value should be a dict {
        'channel_id': {
            'is_selected': True(False),
            'laser': ,
            'filter': ,
            'camera_exposure_time': ,
            'laser_power': ,
            'interval_time':}
        }

        Parameters
        ----------
        setting_dict : dict
            Dictionary containing the values for the experiment.
        """
        self.populate_empty_values()
        self.channel_setting_dict = setting_dict
        prefix = "channel_"
        for channel in setting_dict.keys():
            channel_id = int(channel[len(prefix) :]) - 1
            channel_vals = self.get_vals_by_channel(channel_id)
            if not channel_vals:
                return
            channel_value = setting_dict[channel]
            for name in channel_vals:
                channel_vals[name].set(channel_value[name])

            # validate exposure_time, interval, laser_power
            self.view.exptime_pulldowns[channel_id].trigger_focusout_validation()
            self.view.interval_spins[channel_id].trigger_focusout_validation()
            self.view.laserpower_pulldowns[channel_id].trigger_focusout_validation()

        self.show_verbose_info("channel has been set new value")

    def populate_empty_values(self):
        """Populates the View with empty values.

        If the user changes the number of channels, the new channels need to be populated
        with a default value.
        """
        for i in range(self.num):
            if self.view.laser_pulldowns[i].get() == "":
                self.view.laser_pulldowns[i].set(
                    self.view.laser_pulldowns[i]["values"][0]
                )

            if self.view.filterwheel_pulldowns[i].get() == "":
                self.view.filterwheel_pulldowns[i].set(
                    self.view.filterwheel_pulldowns[i]["values"][0]
                )

            if self.view.exptime_pulldowns[i].get() == "":
                self.view.exptime_pulldowns[i].set(100.0)

            if self.view.laserpower_pulldowns[i].get() == "":
                self.view.laserpower_pulldowns[i].set(10.0)

            if self.view.interval_spins[i].get() == "":
                self.view.interval_spins[i].set(1.0)

    def set_spinbox_range_limits(self, settings):
        """Set the range limits for the spinboxes in the View.

        This function will set the spinbox widget's values of from_, to, step
        according to the settings

        Parameters
        ----------
        settings : dict
            Dictionary containing the range limits for the spinboxes.
        """

        temp_dict = {
            "laser_power": self.view.laserpower_pulldowns,
            "exposure_time": self.view.exptime_pulldowns,
            "interval_time": self.view.interval_spins,
        }
        for k in temp_dict:
            widgets = temp_dict[k]
            for i in range(self.num):
                widgets[i].configure(from_=settings[k]["min"])
                widgets[i].configure(to=settings[k]["max"])
                widgets[i].configure(increment=settings[k]["step"])

    def channel_callback(self, channel_id, widget_name):
        """Callback function for the channel widgets.

        In 'live' mode (when acquire mode is set to 'continuous') and a channel is
        selected, any change of the channel setting will influence devices
        instantly this function will call the central controller to response user's
        request

        Parameters
        ----------
        channel_id : int
            The channel id.
        widget_name : str
            The name of the widget.

        Returns
        -------
        success : bool
            Whether the callback function is executed successfully.

        Examples
        --------
        >>> self.channel_callback(0, "laser")
        """

        channel_vals = self.get_vals_by_channel(channel_id)
        prefix = "channel_"

        def update_setting_dict(setting_dict, widget_name):
            """Update the setting dictionary.

            Parameters
            ----------
            setting_dict : dict
                The setting dictionary.
            widget_name : str
                The name of the widget.

            Returns
            -------
            success : bool
                Whether the setting dictionary is updated successfully.
            """
            if channel_vals[widget_name].get() is None:
                return False

            if widget_name == "laser":
                setting_dict["laser"] = channel_vals["laser"].get()
                setting_dict["laser_index"] = self.get_index(
                    "laser", channel_vals["laser"].get()
                )
            elif widget_name == "filter":
                setting_dict["filter"] = channel_vals["filter"].get()
                setting_dict["filter_position"] = self.get_index(
                    "filter", channel_vals["filter"].get()
                )
            elif widget_name in [
                "laser_power",
                "camera_exposure_time",
                "interval_time",
            ]:
                try:
                    setting_dict[widget_name] = float(channel_vals[widget_name].get())
                except Exception:
                    setting_dict[widget_name] = 0
                    return False
                ref_name = (
                    "exposure_time"
                    if widget_name == "camera_exposure_time"
                    else widget_name
                )
                setting_range = self.parent_controller.parent_controller.configuration[
                    "configuration"
                ]["gui"]["channels"][ref_name]
                if (
                    setting_dict[widget_name] < setting_range["min"]
                    or setting_dict[widget_name] > setting_range["max"]
                ):
                    return False
            else:
                setting_dict[widget_name] = channel_vals[widget_name].get()

            if widget_name == "camera_exposure_time" or widget_name == "is_selected":
                self.parent_controller.execute("recalculate_timepoint")
            return True

        def func(*args):
            """The function to be called when the channel widget is changed.

            Parameters
            ----------
            *args : tuple
                The arguments passed to the callback function.

            Returns
            -------
            success : bool
            """

            if self.in_initialization:
                return

            try:
                if channel_vals[widget_name].get() is None:
                    return
            except tk._tkinter.TclError as e:
                channel_vals[widget_name].set(0)
                # logger.error(f"Tcl Error caught: trying to set position and {e}")
                return

            channel_key = prefix + str(channel_id + 1)
            if channel_key not in self.channel_setting_dict.keys():
                # update self.channel_setting_dict
                setting_dict = self.parent_controller.parent_controller.manager.dict()
                # check whether all the settings are validate
                for name in channel_vals:
                    if not update_setting_dict(setting_dict, name):
                        return
                self.channel_setting_dict[channel_key] = setting_dict
                return

            setting_dict = self.channel_setting_dict[channel_key]

            r = update_setting_dict(setting_dict, widget_name)

            if self.mode == "live":
                # call central controller
                if self.event_id:
                    self.view.after_cancel(self.event_id)
                if r:
                    self.event_id = self.view.after(
                        500,
                        lambda: self.parent_controller.execute(
                            "update_setting", "channel"
                        ),
                    )

            self.show_verbose_info("channel setting has been changed")

        return func

    def get_vals_by_channel(self, index):
        """Get the values of the channel widgets by channel id.

        This function return all the variables according channel_id

        Parameters
        ----------
        index : int
            The channel id.

        Returns
        -------
        dict
            The values of the channel widgets.

        Examples
        --------
        >>> self.get_vals_by_channel(0)
        """
        if index < 0 or index >= self.num:
            return {}
        result = {
            "is_selected": self.view.channel_variables[index],
            "laser": self.view.laser_variables[index],
            "filter": self.view.filterwheel_variables[index],
            "camera_exposure_time": self.view.exptime_variables[index],
            "laser_power": self.view.laserpower_variables[index],
            "interval_time": self.view.interval_variables[index],
            "defocus": self.view.defocus_variables[index],
        }
        return result

    def get_index(self, dropdown_name, value):
        """Get the index of the value in the dropdown list.

        Parameters
        ----------
        dropdown_name : str
            The name of the dropdown list.
        value : str
            The value of the dropdown list.

        Returns
        -------
        int
            The index of the value in the dropdown list.

        Examples
        --------
        >>> self.get_index("laser", "488")
        """
        print("get index", dropdown_name, value)
        if not value:
            return -1
        if dropdown_name == "laser":
            return self.view.laser_pulldowns[0]["values"].index(value)
        elif dropdown_name == "filter":
            return self.view.filterwheel_pulldowns[0]["values"].index(value)
        return -1

    def verify_experiment_values(self):
        """Verify channel settings and return warning info

        Returns
        -------
        string
            Warning info
        """
        selected_channel_num = 0
        setting_range = self.configuration_controller.gui_setting["channels"]
        for channel_key in self.channel_setting_dict.keys():
            setting_dict = self.channel_setting_dict[channel_key]
            if setting_dict["is_selected"]:
                selected_channel_num += 1
                # laser power
                if setting_dict["laser_power"] < setting_range["laser_power"]["min"]:
                    return f"Laser power below configured threshold. Please adjust to meet or exceed the specified minimum in the configuration.yaml({setting_range['laser_power']['min']})."
                elif setting_dict["laser_power"] > setting_range["laser_power"]["max"]:
                    return f"Laser power exceeds configured maximum. Please adjust to meet or be below the specified maximum in the configuration.yaml({setting_range['laser_power']['max']})."
                # exposure time
                if setting_dict["camera_exposure_time"] < setting_range["exposure_time"]["min"]:
                    return f"Exposure time below configured threshold.Please adjust to meet or exceed the specified minimum in the configuration.yaml({setting_range['exposure_time']['min']})."
                elif setting_dict["camera_exposure_time"] > setting_range["exposure_time"]["max"]:
                    return f"Exposure time exceeds configured maximum. Please adjust to meet or be below the specified maximum in the configuration.yaml({setting_range['exposure_time']['max']})"

        if selected_channel_num == 0:
            return "No selected channel!"
        return None



import time
import numpy as np

from pyvcam import pvc
from pyvcam.camera import Camera
from pyvcam import constants

def main():
    pvc.init_pvcam()
    cam = next(Camera.detect_camera())
    cam.open()
    cam.meta_data_enabled = True
    cam.set_roi = (0, 0, 1000, 1000)

    num_frames = 1
    cnt = 0

    cam.start_seq(exp_time=20, num_frames=num_frames)
    while cnt < num_frames:
        frame, fps, frame_count = cam.poll_frame()

        low = np.amin(frame['pixel_data'])
        high = np.amax(frame['pixel_data'])
        average = np.average(frame['pixel_data'])

        print('Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Rate: {:.1f}\tFrame Count: {:.0f}\n'.format(low, high, average, fps, frame_count))
        cnt += 1

        time.sleep(0.05)

    cam.finish()
    cam.close()
    pvc.uninit_pvcam()

    print('Total frames: {}\n'.format(cnt))


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk

# Third Party Imports
from matplotlib.pyplot import subplots
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

# Local Imports
from navigate.view.custom_widgets.popup import PopUp


class CameraMapSettingPopup(PopUp):
    """Popup to create and visualize camera offset and variance map generation."""

    def __init__(
        self,
        root,
        name="Camera Map Settings",
        size="+320+180",
        top=True,
        transient=True,
        *args,
        **kwargs
    ):
        """Initialize the CameraMapSettingPopup.

        Parameters
        ----------
        root : tk.Tk
            Root window.
        name : str, optional
            Name of the popup, by default "Camera Map Settings".
        size : str, optional
            Size of the popup, by default "+320+180".
        top : bool, optional
            Whether the popup should be on top, by default True.
        transient : bool, optional
            Whether the popup should be transient, by default True.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.
        """
        super().__init__(root, name, size, top, transient, *args, **kwargs)

        #: dict: Dictionary of input widgets.
        self.inputs = {}

        title = ttk.Label(self.content_frame, text="File: ", padding=(2, 5, 0, 0))
        title.grid(row=0, column=0, sticky=(tk.NSEW))

        #: tk.StringVar: File name.
        self.file_name = tk.StringVar()
        self.inputs["file_name"] = tk.Entry(
            self.content_frame, textvariable=self.file_name
        )
        self.inputs["file_name"].grid(
            row=0, column=1, sticky=tk.NSEW, padx=(0, 5), pady=(15, 0)
        )

        #: ttk.Button: Open file button.
        self.open_btn = ttk.Button(self.content_frame, text="Open")
        self.open_btn.grid(row=0, column=2, pady=(0, 10))

        title = ttk.Label(self.content_frame, text="Camera: ", padding=(2, 5, 0, 0))
        #: tk.StringVar: Camera name.
        self.camera = tk.StringVar()
        title.grid(row=0, column=3, sticky=(tk.NSEW))
        self.inputs["camera"] = ttk.OptionMenu(self.content_frame, self.camera)
        self.inputs["camera"].grid(
            row=0, column=4, sticky=tk.NSEW, padx=(0, 5), pady=(15, 0)
        )
        #: ttk.Button: Create maps button.
        self.map_btn = ttk.Button(self.content_frame, text="Create maps")
        self.map_btn.grid(row=0, column=5, pady=(0, 10))

        # Plot
        self.fig, self.axs = subplots(1, 2, figsize=(5, 5))
        canvas = FigureCanvasTkAgg(self.fig, master=self.content_frame)
        canvas.draw()
        canvas.get_tk_widget().grid(
            row=1, column=0, columnspan=6, sticky=tk.NSEW, padx=(5, 5), pady=(5, 5)
        )

    def get_widgets(self):
        """Get the input widgets.

        Returns
        -------
        dict
            Dictionary of input widgets.
        """
        return self.inputs



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging

# Third Party Library Imports

# Local Library Imports


class PerformanceFilter(logging.Filter):
    """
    A custom logging filter to exclude performance messages.

    This filter is designed to be used with the Python logging module. It includes
    log records whose messages do not start with "Performance" or "Spec."
    """

    def filter(self, record):
        """Filter out performance messages

        Parameters
        ----------
        record : logging.LogRecord
            The log record to filter

        Returns
        -------
        bool
            True if the record should be logged, False otherwise
        """
        # Checking if log message should be sent to performance.log
        # based on if it starts with Performance or Spec
        if record.getMessage().startswith("Performance"):
            return True
        if record.getMessage().startswith("Spec"):
            return True

        return False


class NonPerfFilter(logging.Filter):
    """
    A custom logging filter to exclude non-performance messages.

    This filter is designed to be used with the Python logging module. It excludes
    log records whose messages do not start with "Performance" or "Spec."
    """

    def filter(self, record):
        """Filter out non-performance messages

        Parameters
        ----------
        record : logging.LogRecord
            The log record to filter

        Returns
        -------
        bool
            True if the record should be logged, False otherwise
        """
        # Making sure performance data only goes to performance.log

        if record.getMessage().startswith("Performance"):
            return False
        if record.getMessage().startswith("Spec"):
            return False
        return True



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
import platform
import os
import tkinter as tk
from tkinter import messagebox, filedialog
import subprocess
import webbrowser

# Third Party Imports

# Local Imports
from navigate.view.popups.ilastik_setting_popup import ilastik_setting_popup
from navigate.view.popups.autofocus_setting_popup import AutofocusPopup
from navigate.view.popups.adaptiveoptics_popup import AdaptiveOpticsPopup
from navigate.view.popups.camera_map_setting_popup import CameraMapSettingPopup
from navigate.view.popups.waveform_parameter_popup_window import (
    WaveformParameterPopupWindow,
)
from navigate.view.popups.feature_list_popup import FeatureListPopup
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.controller.sub_controllers import (
    AutofocusPopupController,
    IlastikPopupController,
    CameraMapSettingPopupController,
    WaveformPopupController,
    MicroscopePopupController,
    FeaturePopupController,
    FeatureAdvancedSettingController,
    AdaptiveOpticsPopupController,
    UninstallPluginController,
)
from navigate.tools.file_functions import save_yaml_file, load_yaml_file
from navigate.tools.decorators import FeatureList
from navigate.tools.common_functions import load_module_from_file
from navigate.config.config import get_navigate_path


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class FakeEvent:
    """Fake event class for keyboard shortcuts"""

    def __init__(self, char=None, keysym=None):
        """Initialize FakeEvent

        Parameters
        ----------
        char: str
            The character that was pressed.
        keysym: str
            The key that was pressed.
        """
        #: str: The character that was pressed.
        self.char = char
        #: str: The key that was pressed.
        self.keysym = keysym
        #: int: The state of the keyboard.
        self.state = 0


class MenuController(GUIController):
    """Menu controller class."""

    def __init__(self, view, parent_controller=None):
        """Initialize MenuController

        Parameters
        ----------
        view: class
            The view class.
        parent_controller
            The parent controller.
        """
        super().__init__(view, parent_controller)
        #: Controller: The parent controller.
        self.parent_controller = parent_controller
        #: tk.canvas: The view class.
        self.view = view
        #: tkinter.StringVar: Resolution value.
        self.resolution_value = tk.StringVar()
        #: tkinter.IntVar: Feature id value.
        self.feature_id_val = tk.IntVar()
        #: tkinter.IntVar: Disable stage limits.
        self.disable_stage_limits = tk.IntVar()
        #: FakeEvent: Fake event.
        self.fake_event = None
        #: list: List of feature list names.
        self.feature_list_names = []
        #: int: System feature list count.
        self.system_feature_list_count = 0
        #: int: Feature list count.
        self.feature_list_count = 0
        #: str: Feature list file name.
        self.feature_list_file_name = "feature_lists.yaml"

    def initialize_menus(self):
        """Initialize menus
        This function defines all the menus in the menubar

        Each menu item is initialized as a dictionary entry that is associated with
        a list that provides the following parameters:

        Menu item name: name of the menu item. If the name is specified as
        add_separator, then a separator is added to the menu.

        List of parameters:
            Type of entry: standard, checkbutton, radiobutton, cascade
            Function: function to be called when menu item is selected
            Accelerator: keyboard shortcut
            Bindings: keyboard shortcut bindings for Windows
            Bindings: keyboard shortcut bindings for Mac

        Example:
            "Acquire Data": [
                "standard",
                self.acquire_data,
                "Ctrl+Enter",
                "<Control-Return>",
                "<Control_L-Return>",
            ]

        Returns
        -------
        configuration_controller : class
            Camera view sub-controller.

        """

        # File Menu
        file_menu = {
            self.view.menubar.menu_file: {
                "New Experiment": [
                    "standard",
                    self.new_experiment,
                    "Ctrl+Shift+N",
                    "<Control-N>",
                    "<Control_L-N>",
                ],
                "Load Experiment": [
                    "standard",
                    self.load_experiment,
                    "Ctrl+Shift+O",
                    "<Control-O>",
                    "<Control_L-O>",
                ],
                "Save Experiment": [
                    "standard",
                    self.save_experiment,
                    "Ctrl+Shift+S",
                    "<Control-S>",
                    "<Control_L-S>",
                ],
                "add_separator": [None],
                "Toggle Save Data": [
                    "standard",
                    self.toggle_save,
                    "Ctrl+s",
                    "<Control-s>",
                    "<Control_L-s>",
                ],
                "Acquire Data": [
                    "standard",
                    self.acquire_data,
                    "Ctrl+Enter",
                    "<Control-Return>",
                    "<Control_L-Return>",
                ],
                "Load Images": ["standard", self.load_images, None, None, None],
                "Unload Images": [
                    "standard",
                    lambda: self.parent_controller.model.load_images(None),
                    None,
                    None,
                    None,
                ],
                "add_separator_1": [None, None, None, None, None],
                "Open Log Files": ["standard", self.open_log_files, None, None, None],
                "Open Configuration Files": [
                    "standard",
                    self.open_configuration_files,
                    None,
                    None,
                    None,
                ],
            }
        }
        self.populate_menu(file_menu)

        # Stage Control Menu
        # Most bindings are implemented in the keystroke_controller.
        # Accelerators added here to communicate them to users. Could move those key
        # bindings here? Not sure...
        stage_control_menu = {
            self.view.menubar.menu_multi_positions: {
                "Move Up": [
                    "standard",
                    lambda *args: self.stage_movement("w"),
                    "w",
                    "<Key-w>",
                    "<Key-w>",
                ],
                "Move Down": [
                    "standard",
                    lambda *args: self.stage_movement("s"),
                    "s",
                    "<Key-s>",
                    "<Key-s>",
                ],
                "Move Left": [
                    "standard",
                    lambda *args: self.stage_movement("a"),
                    "a",
                    "<Key-a>",
                    "<Key-a>",
                ],
                "Move Right": [
                    "standard",
                    lambda *args: self.stage_movement("d"),
                    "d",
                    "<Key-d>",
                    "<Key-d>",
                ],
                "Move In": ["standard", self.not_implemented, None, None, None],
                "Move Out": ["standard", self.not_implemented, None, None, None],
                "Move Focus Up": ["standard", self.not_implemented, None, None, None],
                "Move Focus Down": ["standard", self.not_implemented, None, None, None],
                "Rotate Clockwise": [
                    "standard",
                    self.not_implemented,
                    None,
                    None,
                    None,
                ],
                "Rotate Counter-Clockwise": [
                    "standard",
                    self.not_implemented,
                    None,
                    None,
                    None,
                ],
                "add_separator": ["standard", None, None, None, None],
                "Launch Tiling Wizard": [
                    "standard",
                    self.parent_controller.channels_tab_controller.launch_tiling_wizard,
                    None,
                    None,
                    None,
                ],
                "Load Positions": [
                    "standard",
                    self.parent_controller.multiposition_tab_controller.load_positions,
                    None,
                    None,
                    None,
                ],
                "Export Positions": [
                    "standard",
                    self.parent_controller.multiposition_tab_controller.export_positions,  # noqa: E501
                    None,
                    None,
                    None,
                ],
                "Append Current Position": [
                    "standard",
                    self.parent_controller.multiposition_tab_controller.add_stage_position,  # noqa: E501
                    None,
                    None,
                    None,
                ],
                "add_separator_1": [None, None, None, None, None],
            },
        }
        self.populate_menu(stage_control_menu)
        self.view.menubar.menu_multi_positions.add_radiobutton(
            label="Disable Stage Limits",
            value=1,
            command=self.toggle_stage_limits,
            variable=self.disable_stage_limits,
        )
        self.view.menubar.menu_multi_positions.add_radiobutton(
            label="Enable Stage Limits",
            value=0,
            command=self.toggle_stage_limits,
            variable=self.disable_stage_limits,
        )

        # autofocus menu
        autofocus_menu = {
            self.view.menubar.menu_autofocus: {
                "Perform Autofocus": [
                    "standard",
                    lambda *args: self.parent_controller.execute("autofocus"),
                    "Ctrl+Shift+A",
                    "<Control-A>",
                    "<Control_L-A>",
                ],
                "Autofocus Settings": [
                    "standard",
                    self.popup_autofocus_setting,
                    "Ctrl+Alt+Shift+A",
                    "<Control-Alt-A>",
                    "<Command-Alt-Key-A>",
                ],
            }
        }
        self.populate_menu(autofocus_menu)

        # Window menu
        windows_menu = {
            self.view.menubar.menu_window: {
                "Channel Settings": [
                    "standard",
                    lambda event: self.switch_tabs(1),
                    "Ctrl+1",
                    "<Control-Key-1>",
                    "<Control_L-Key-1",
                ],
                "Camera Settings": [
                    "standard",
                    lambda event: self.switch_tabs(2),
                    "Ctrl+2",
                    "<Control-Key-2>",
                    "<Control_L-Key-2",
                ],
                "Stage Control": [
                    "standard",
                    lambda event: self.switch_tabs(3),
                    "Ctrl+3",
                    "<Control-Key-3>",
                    "<Control_L-Key-3",
                ],
                "Multiposition Table": [
                    "standard",
                    lambda event: self.switch_tabs(4),
                    "Ctrl+4",
                    "<Control-Key-4>",
                    "<Control_L-Key-4",
                ],
                "add_separator": ["standard", None, None, None, None],
                "Popout Camera Display": [
                    "standard",
                    self.not_implemented,
                    None,
                    None,
                    None,
                ],
                "Help": ["standard", self.popup_help, None, None, None],
            }
        }
        self.populate_menu(windows_menu)

        # Zoom menu
        for microscope_name in self.parent_controller.configuration["configuration"][
            "microscopes"
        ].keys():
            zoom_positions = self.parent_controller.configuration["configuration"][
                "microscopes"
            ][microscope_name]["zoom"]["position"]
            if len(zoom_positions) > 1:
                sub_menu = tk.Menu(self.view.menubar.menu_resolution)
                self.view.menubar.menu_resolution.add_cascade(
                    menu=sub_menu, label=microscope_name
                )
                for res in zoom_positions.keys():
                    sub_menu.add_radiobutton(
                        label=res,
                        variable=self.resolution_value,
                        value=f"{microscope_name} {res}",
                    )
            else:
                self.view.menubar.menu_resolution.add_radiobutton(
                    label=microscope_name,
                    variable=self.resolution_value,
                    value=f"{microscope_name} {zoom_positions.keys()[0]}",
                )
        self.resolution_value.trace_add(
            "write",
            lambda *args: self.parent_controller.execute(
                "resolution", self.resolution_value.get()
            ),
        )

        configuration_dict = {
            self.view.menubar.menu_resolution: {
                "add_separator": [None, None, None, None, None],
                "Waveform Parameters": [
                    "standard",
                    self.popup_waveform_setting,
                    None,
                    None,
                    None,
                ],
                "Configure Microscope": [
                    "standard",
                    self.popup_microscope_setting,
                    None,
                    None,
                    None,
                ],
            }
        }
        self.populate_menu(configuration_dict)

        # plugins
        self.view.menubar.menu_plugins.add_command(
            label="Install Plugin", command=self.install_plugin
        )
        self.view.menubar.menu_plugins.add_command(
            label="Uninstall Plugins", command=self.popup_uninstall_plugin
        )
        self.view.menubar.menu_plugins.add_separator()

        # add-on features
        self.feature_list_names = [
            "None",
            "Switch Resolution",
            "Z Stack Acquisition",
            "Threshold",
            "Ilastik Segmentation",
            "Volume Search",
            "Time Series",
            "Decoupled Focus Stage Multiposition",
            "Remove Empty Tiles",
        ]
        self.feature_list_count = len(self.feature_list_names)
        self.system_feature_list_count = self.feature_list_count

        for i in range(self.feature_list_count):
            self.view.menubar.menu_features.add_radiobutton(
                label=self.feature_list_names[i], variable=self.feature_id_val, value=i
            )
        self.feature_id_val.trace_add(
            "write",
            lambda *args: self.parent_controller.execute(
                "load_feature", self.feature_id_val.get()
            ),
        )

        # add adaptive optics as standalone pop-up for now
        # (if mirror is real, otherwise don't)
        microscope_name = self.parent_controller.configuration["experiment"][
            "MicroscopeState"
        ]["microscope_name"]
        scope = self.parent_controller.configuration["configuration"]["microscopes"][
            microscope_name
        ]
        if (
            scope.get("mirror") is not None
            and "synthetic" not in scope["mirror"]["hardware"]["type"].lower()
        ):
            self.view.menubar.menu_features.add_separator()
            self.view.menubar.menu_features.add_command(
                label="Adaptive Optics", command=self.popup_adaptiveoptics
            )

        self.view.menubar.menu_features.add_separator()

        self.view.menubar.menu_features.add_command(
            label="Ilastik Settings", command=self.popup_ilastik_setting
        )
        # disable ilastik menu
        self.view.menubar.menu_features.entryconfig(
            "Ilastik Segmentation", state="disabled"
        )
        self.view.menubar.menu_features.add_command(
            label="Camera offset and variance maps",
            command=self.popup_camera_map_setting,
        )
        self.view.menubar.menu_features.add_command(
            label="Load Customized Feature List", command=self.load_feature_list
        )
        self.view.menubar.menu_features.add_command(
            label="Add Customized Feature List", command=self.popup_feature_list_setting
        )
        self.view.menubar.menu_features.add_command(
            label="Delete Selected Feature List", command=self.delete_feature_list
        )
        self.view.menubar.menu_features.add_command(
            label="Advanced Setting", command=self.popup_feature_advanced_setting
        )
        self.view.menubar.menu_features.add_separator()

        # add feature lists from previous loaded ones
        feature_lists_path = get_navigate_path() + "/feature_lists"
        if not os.path.exists(feature_lists_path):
            os.makedirs(feature_lists_path)
            return
        # get __sequence.yml
        feature_records = load_yaml_file(f"{feature_lists_path}/__sequence.yml")
        if not feature_records:
            return

        for feature in feature_records:
            self.view.menubar.menu_features.add_radiobutton(
                label=feature["feature_list_name"],
                variable=self.feature_id_val,
                value=self.feature_list_count,
            )
            self.feature_list_names.append(feature["feature_list_name"])
            self.feature_list_count += 1

        # Note: Any menu items added below this return statement will not
        # be populated if feature_records does not exist.

    def toggle_save(self, *args):
        """Toggle save button

        Parameters
        ----------
        args:
            could be tkinter event(Key press event)

        """
        save_data = (
            self.view.settings.channels_tab.stack_timepoint_frame.save_data.get()
        )

        self.parent_controller.channels_tab_controller.timepoint_vals["is_save"].set(
            not save_data
        )
        self.parent_controller.channels_tab_controller.update_save_setting()

    def open_folder(self, path):
        """Open folder in file explorer.

        Parameters
        ----------
        path : str
            Path to folder.
        """
        try:
            if platform.system() == "Darwin":  # macOS
                subprocess.check_call(["open", "--", path])
            elif platform.system() == "Windows":  # Windows
                subprocess.check_call(["explorer", path])
            else:
                print("Unsupported platform.")
        except subprocess.CalledProcessError:
            pass

    def open_log_files(self):
        """Open log files folder."""
        path = os.path.join(get_navigate_path(), "logs")
        self.open_folder(path)

    def open_configuration_files(self):
        """Open configuration files folder."""
        path = os.path.join(get_navigate_path(), "config")
        self.open_folder(path)

    def populate_menu(self, menu_dict):
        """Populate the menus from a dictionary.

        Parameters
        ----------
        menu_dict : dict
            menu_dict = {
                Menu object: {
                    "Menu String Entry": [
                        entry_type (standard, radio),
                        Command,
                        Accelerator,
                        Windows Keystroke,
                        Apple Keystroke",
            ],
            ....

        """
        for menu in menu_dict:
            menu_items = menu_dict[menu]
            for label in menu_items:
                if "add_separator" in label:
                    menu.add_separator()
                else:
                    if "standard" in menu_items[label][0]:
                        if menu_items[label][1] is None:
                            # Command not passed, accelerator provided for
                            # informational purposes only.
                            menu.add_command(
                                label=label, accelerator=menu_items[label][2]
                            )
                        else:
                            # If the command is provided, it is assumed that you
                            # should also bind that command to the accelerator.
                            menu.add_command(
                                label=label,
                                command=menu_items[label][1],
                                accelerator=menu_items[label][2],
                            )
                            if platform.platform() == "Darwin":
                                # Account for OS specific keystrokes
                                menu.bind_all(
                                    menu_items[label][4], menu_items[label][1]
                                )
                            else:
                                menu.bind_all(
                                    menu_items[label][3], menu_items[label][1]
                                )
                    elif "radio" in menu_items[label][0]:
                        if menu_items[label][1] is None:
                            # Command not passed, accelerator provided for
                            # informational purposes only.
                            menu.add_radiobutton(
                                label=label, accelerator=menu_items[label][2]
                            )
                        else:
                            # If the command is provided, it is assumed that you
                            # should also bind that command to the accelerator.
                            menu.add_radiobutton(
                                label=label,
                                command=menu_items[label][0],
                                accelerator=menu_items[label][1],
                            )
                            if platform.platform() == "Darwin":
                                menu.bind_all(
                                    menu_items[label][4], menu_items[label][1]
                                )
                            else:
                                menu.bind_all(
                                    menu_items[label][3], menu_items[label][1]
                                )

    def new_experiment(self, *args):
        """Create a new experiment file."""
        self.parent_controller.populate_experiment_setting(
            self.parent_controller.default_experiment_file
        )

    def load_experiment(self, *args):
        """Load an experiment file."""
        filename = filedialog.askopenfilename(
            defaultextension=".yml", filetypes=[("Yaml files", "*.yml *.yaml")]
        )
        if not filename:
            return
        self.parent_controller.populate_experiment_setting(filename)

    def save_experiment(self, *args):
        """Save an experiment file.

        Updates model.experiment and saves it to file.
        """
        if not self.parent_controller.update_experiment_setting():
            messagebox.showerror(
                title="Warning",
                message="Incorrect/missing settings. "
                "Cannot save current experiment file.",
            )
            return
        filename = filedialog.asksaveasfilename(
            defaultextension=".yml", filetypes=[("Yaml file", "*.yml *.yaml")]
        )
        if not filename:
            return
        save_yaml_file("", self.parent_controller.configuration["experiment"], filename)

    def load_images(self):
        """Load images from a file."""
        filenames = filedialog.askopenfilenames(
            defaultextension=".tif", filetypes=[("tiff files", "*.tif *.tiff")]
        )
        if not filenames:
            return
        self.parent_controller.model.load_images(filenames)

    def popup_camera_map_setting(self):
        """Pop up the Camera Map setting window."""
        if hasattr(self.parent_controller, "camera_map_popup_controller"):
            self.parent_controller.camera_map_popup_controller.showup()
            return
        map_popup = CameraMapSettingPopup(self.view)
        self.parent_controller.camera_map_popup_controller = (
            CameraMapSettingPopupController(map_popup, self.parent_controller)
        )

    def popup_adaptiveoptics(self):
        """Pop up the Adaptive Optics setting window."""
        if hasattr(self.parent_controller, "adaptiveoptics_popup_controller"):
            self.parent_controller.ao_popup_controller.showup()
            return
        ao_popup = AdaptiveOpticsPopup(self.view)
        self.parent_controller.ao_popup_controller = AdaptiveOpticsPopupController(
            ao_popup, self.parent_controller
        )

    def popup_ilastik_setting(self):
        """Pop up the Ilastik setting window."""
        ilastik_popup_window = ilastik_setting_popup(self.view)
        ilastik_url = self.parent_controller.configuration["rest_api_config"][
            "Ilastik"
        ]["url"]
        if hasattr(self.parent_controller, "ilastik_controller"):
            self.parent_controller.ilastik_controller.showup(ilastik_popup_window)
        else:
            self.parent_controller.ilastik_controller = IlastikPopupController(
                ilastik_popup_window, self.parent_controller, ilastik_url
            )

    def popup_help(self):
        """Open a web browser to the Navigate documentation."""
        webbrowser.open_new_tab("https://thedeanlab.github.io/navigate/")

    def toggle_stage_limits(self, *args):
        """Toggle stage limits."""
        if self.disable_stage_limits.get() == 1:
            self.parent_controller.configuration["experiment"]["StageParameters"][
                "limits"
            ] = False
            logger.debug("Disabling stage limits")
            self.parent_controller.execute("stage_limits", False)
        else:
            self.parent_controller.configuration["experiment"]["StageParameters"][
                "limits"
            ] = True
            logger.debug("Enabling stage limits")
            self.parent_controller.execute("stage_limits", True)

    def popup_autofocus_setting(self, *args):
        """Pop up the Autofocus setting window."""
        if hasattr(self.parent_controller, "af_popup_controller"):
            self.parent_controller.af_popup_controller.showup()
            return
        af_popup = AutofocusPopup(self.view)
        self.parent_controller.af_popup_controller = AutofocusPopupController(
            af_popup, self.parent_controller
        )

    def popup_waveform_setting(self):
        if hasattr(self.parent_controller, "waveform_popup_controller"):
            self.parent_controller.waveform_popup_controller.showup()
            return
        waveform_constants_popup = WaveformParameterPopupWindow(
            self.view, self.parent_controller.configuration_controller
        )
        waveform_popup_controller = WaveformPopupController(
            waveform_constants_popup,
            self.parent_controller,
            self.parent_controller.waveform_constants_path,
        )
        waveform_popup_controller.populate_experiment_values()
        self.parent_controller.waveform_popup_controller = waveform_popup_controller

    def popup_microscope_setting(self):
        """Pop up the microscope setting window."""
        if hasattr(self.parent_controller, "microscope_popup_controller"):
            self.parent_controller.microscope_popup_controller.showup()
            return
        microscope_info = self.parent_controller.model.get_microscope_info()
        self.parent_controller.microscope_popup_controller = MicroscopePopupController(
            self.view, self.parent_controller, microscope_info
        )

    def acquire_data(self, *args):
        """Acquire data/Stop acquiring data."""
        self.parent_controller.acquire_bar_controller.launch_popup_window()

    def not_implemented(self, *args):
        """Not implemented."""
        print("Not implemented")

    def stage_movement(self, char):
        """Stage movement.

        Should not be run if we are in a validated combobox, or a validate entry.

        Parameters
        ----------
        char: str
            The character that was pressed.
        """
        try:
            focus = self.parent_controller.view.focus_get()
            if hasattr(focus, "widgetName"):
                freeze_in = ["ttk::entry", "ttk::combobox", "text", "ttk::spinbox"]
                if focus.widgetName in freeze_in:
                    return
            self.fake_event = FakeEvent(char=char)
            self.parent_controller.stage_controller.stage_key_press(self.fake_event)
        except KeyError:
            # Avoids KeyError if the user is in a popdown menu.
            pass

    def switch_tabs(self, tab):
        """Switch tabs."""
        self.parent_controller.view.settings.select(tab - 1)

    def popup_feature_list_setting(self):
        """Show feature list popup window"""
        feature_list_popup = FeatureListPopup(self.view, title="Add New Feature List")
        self.parent_controller.features_popup_controller = FeaturePopupController(
            feature_list_popup, self.parent_controller
        )

    def load_feature_list(self):
        """Load feature lists from a python file"""
        filename = filedialog.askopenfilename(
            defaultextension=".py", filetypes=[("Python files", "*.py")]
        )
        if not filename:
            return
        module = load_module_from_file(filename[filename.rindex("/") + 1 :], filename)
        features = [
            f for f in dir(module) if isinstance(getattr(module, f), FeatureList)
        ]
        feature_lists_path = get_navigate_path() + "/feature_lists"
        feature_list_files = [
            temp
            for temp in os.listdir(feature_lists_path)
            if (temp.endswith(".yml") or temp.endswith(".yaml"))
            and os.path.isfile(os.path.join(feature_lists_path, temp))
        ]
        feature_records = load_yaml_file(f"{feature_lists_path}/__sequence.yml")
        if not feature_records:
            feature_records = []
        added_features = []
        for name in features:
            feature = getattr(module, name)
            feature_list_name = feature.feature_list_name
            if (
                f"{feature_list_name}.yml" in feature_list_files
                or f"{feature_list_name}.yaml" in feature_list_files
            ):
                print(
                    "There is already one feature list named as",
                    feature_list_name,
                    "The new one isn't loaded!",
                )
                continue
            self.view.menubar.menu_features.add_radiobutton(
                label=feature_list_name,
                variable=self.feature_id_val,
                value=self.feature_list_count,
            )
            save_yaml_file(
                feature_lists_path,
                {
                    "module_name": name,
                    "feature_list_name": feature_list_name,
                    "filename": filename,
                },
                f"{'_'.join(feature_list_name.split(' '))}.yml",
            )

            feature_records.append(
                {
                    "feature_list_name": feature_list_name,
                    "yaml_file_name": "_".join(feature_list_name.split(" ")) + ".yml",
                }
            )
            self.feature_list_names.append(feature_list_name)
            self.feature_list_count += 1
            added_features.append(name)

        save_yaml_file(feature_lists_path, feature_records, "__sequence.yml")
        # tell model to add feature lists
        self.parent_controller.model.load_feature_list_from_file(
            filename, added_features
        )

    def add_feature_list(self, feature_list_name, feature_list_str):
        """Add feature list to the software and system yaml files

        Parameters
        ----------
        feature_list_name: str
            feature list name
        feature_list_str: str
            string of a feature list

        Returns
        -------
        result : bool
            True: add feature list successfully
            False: failed
        """
        feature_lists_path = get_navigate_path() + "/feature_lists"
        if os.path.exists(f"{feature_lists_path}/{'_'.join(feature_list_name)}.yml"):
            return False
        self.view.menubar.menu_features.add_radiobutton(
            label=feature_list_name,
            variable=self.feature_id_val,
            value=self.feature_list_count,
        )
        self.feature_list_names.append(feature_list_name)
        self.feature_list_count += 1
        save_yaml_file(
            feature_lists_path,
            {
                "module_name": None,
                "feature_list_name": feature_list_name,
                "feature_list": feature_list_str,
            },
            f"{'_'.join(feature_list_name.split(' '))}.yml",
        )
        feature_records = load_yaml_file(f"{feature_lists_path}/__sequence.yml")
        feature_records.append(
            {
                "feature_list_name": feature_list_name,
                "yaml_file_name": "_".join(feature_list_name.split(" ")) + ".yml",
            }
        )
        # tell model to add feature lists
        self.parent_controller.model.load_feature_list_from_str(feature_list_str)
        # save feature records
        save_yaml_file(feature_lists_path, feature_records, "__sequence.yml")
        return True

    def delete_feature_list(self):
        """Delete a selected customized feature list from the software and system
        yaml file"""
        feature_id = self.feature_id_val.get()
        if feature_id < self.system_feature_list_count:
            messagebox.showerror(
                title="Feature List Error",
                message="Can't delete system feature list or you haven't select any "
                "feature list",
            )
            return

        feature_list_name = self.feature_list_names[feature_id]
        self.view.menubar.menu_features.delete(feature_list_name)

        # remove from yaml file
        feature_lists_path = get_navigate_path() + "/feature_lists"
        feature_records = load_yaml_file(f"{feature_lists_path}/__sequence.yml")
        temp = feature_records[feature_id - self.system_feature_list_count]
        os.remove(f"{feature_lists_path}/{temp['yaml_file_name']}")

        del feature_records[feature_id - self.system_feature_list_count]
        save_yaml_file(feature_lists_path, feature_records, "__sequence.yml")

    def popup_feature_advanced_setting(self):
        """Show feature advanced setting window"""
        self.parent_controller.feature_advanced_setting_controller = (
            FeatureAdvancedSettingController(self.view, self.parent_controller)
        )

    def install_plugin(self, *args):
        """Install a plugin"""
        folder_path = filedialog.askdirectory()
        if not folder_path:
            return
        if os.path.exists(os.path.join(folder_path, "plugin_config.yml")):
            plugin_config = load_yaml_file(
                os.path.join(folder_path, "plugin_config.yml")
            )
            plugins_dict = load_yaml_file(
                os.path.join(get_navigate_path(), "config", "plugins_config.yml")
            )
            plugin_name = plugin_config["name"]
            if plugin_name in plugins_dict:
                messagebox.showwarning(
                    title="Warning",
                    message=f"Plugin {plugin_name} already exists,"
                    "Cannot install the selected one.",
                )
                return
            else:
                plugins_dict[plugin_name] = folder_path
                save_yaml_file(
                    os.path.join(get_navigate_path(), "config"),
                    plugins_dict,
                    "plugins_config.yml",
                )
                messagebox.showwarning(
                    title="Plugin",
                    message=f"Plugin {plugin_name} is installed!"
                    "Please restart Navigate!",
                )

    def popup_uninstall_plugin(self, *args):
        """Uninstall plugin"""
        if hasattr(self, "uninstall_plugin_controller"):
            self.uninstall_plugin_controller.showup()
            return
        self.uninstall_plugin_controller = UninstallPluginController(self.view, self)
